from .tools import *  # noqa
from .viewers import *  # noqa
from .image_viewer_creator import *  # noqa
from .parsers import *  # noqa
from .coords_info import *  # noqa
from .orientation import *  # noqa
from .compass import *  # noqa
from .aper_phot_simple import *  # noqa
from .line_profile_xy import *  # noqa
from .catalogs import *  # noqa
from .footprints import *  # noqa



from glue.core.message import EditSubsetMessage
from glue.core.edit_subset_mode import (AndMode, AndNotMode, OrMode,
                                        ReplaceMode, XorMode)
from glue_jupyter.widgets.subset_mode_vuetify import SelectionModeMenu
from glue_jupyter.widgets.subset_select_vuetify import SubsetSelect
from traitlets import Int, List

from jdaviz.core.registries import tool_registry
from jdaviz.core.template_mixin import TemplateMixin

__all__ = ['SubsetAppBarWidget']

SUBSET_MODES = {
    'replace': ReplaceMode,
    'add': OrMode,
    'and': AndMode,
    'xor': XorMode,
    'remove': AndNotMode
}


@tool_registry('g-subset-tools')
class SubsetAppBarWidget(TemplateMixin):
    template_file = __file__, "subset_widget.vue"
    select = List([]).tag(sync=True)
    subset_mode = Int(0).tag(sync=True)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.components = {
            'g-subset-select': SubsetSelect(session=self.session),
            'g-subset-mode': SelectionModeMenu(session=self.session)
        }

        self.hub.subscribe(self, EditSubsetMessage,
                           handler=self._subset_edit_event)

    def _subset_edit_event(self, msg):
        # NOTE: here we'll assume that only a single subset is selected (glue supports
        # multiple subsets being selected, but jdaviz doesn't)
        if not (len(msg.subset)):
            # then changed to "Create New", we want to revert the mode to ReplaceMode
            self.session.edit_subset_mode.mode = ReplaceMode



import numpy as np
from astropy.coordinates import Angle
from astropy.nddata import NDData
from astropy.tests.helper import assert_quantity_allclose
from numpy.testing import assert_allclose
from regions import PixCoord, CirclePixelRegion, RectanglePixelRegion, EllipsePixelRegion

from jdaviz.configs.imviz.tests.utils import BaseImviz_WCS_WCS, BaseImviz_WCS_GWCS


class TestDeleteData(BaseImviz_WCS_WCS):

    def test_reparent_str(self):
        for subset in self.imviz.app.data_collection.subset_groups:
            self.imviz.app._reparent_subsets(
                subset.subset_state.xatt.parent.label,
                "has_wcs_1[SCI,1]"
            )

    def test_delete_with_subset_wcs(self):
        # Add a third dataset to test relinking
        arr = np.ones((10, 10))

        # First data with WCS, same as the one in BaseImviz_WCS_NoWCS.
        hdu3 = NDData(arr, wcs=self.wcs_1)
        self.imviz.load_data(hdu3, data_label='has_wcs_3')

        self.imviz.link_data(align_by='wcs', wcs_fallback_scheme=None)

        # Add a subset
        reg = CirclePixelRegion(PixCoord(2, 2), 3).to_sky(self.wcs_1)
        self.imviz.plugins['Subset Tools'].import_region(reg)

        self.imviz.plugins['Subset Tools'].combination_mode = 'new'
        reg = RectanglePixelRegion(PixCoord(1, 1), 2, 2).to_sky(self.wcs_1)
        self.imviz.plugins['Subset Tools'].import_region(reg)

        assert len(self.imviz.app.data_collection.subset_groups) == 2

        # by default the parent will be the reference data layer, which
        # is the "Default Orientation" WCS-only layer since Imviz is
        # WCS-linked here. Let's reparent to a layer that we can
        # delete to test the machinery:
        for subset in self.imviz.app.data_collection.subset_groups:
            self.imviz.app._reparent_subsets(
                subset.subset_state.xatt.parent.label,
                "has_wcs_1[SCI,1]"
            )

        subset1 = self.imviz.app.data_collection.subset_groups[0]
        subset2 = self.imviz.app.data_collection.subset_groups[1]
        assert subset1.subset_state.xatt.parent.label == "has_wcs_1[SCI,1]"
        assert_allclose(subset1.subset_state.center(), (2, 2))

        assert subset2.subset_state.xatt.parent.label == "has_wcs_1[SCI,1]"
        assert_allclose(subset2.subset_state.roi.xmin, 0, atol=1e-6)
        assert_allclose(subset2.subset_state.roi.ymin, 0, atol=1e-6)
        assert_allclose(subset2.subset_state.roi.xmax, 2)
        assert_allclose(subset2.subset_state.roi.ymax, 2)

        # We have to remove the data from the viewer before deleting the data from the app.
        self.imviz.app.remove_data_from_viewer("imviz-0", "has_wcs_1[SCI,1]")
        self.imviz.app.data_item_remove("has_wcs_1[SCI,1]")

        # Make sure we re-linked images 2 and 3 (plus WCS-only reference data layer)
        assert len(self.imviz.app.data_collection.external_links) == 2

        # FIXME: 0.25 offset introduced by fake WCS layer, see
        # https://jira.stsci.edu/browse/JDAT-4256

        # Check that the reparenting and coordinate recalculations happened
        assert subset1.subset_state.xatt.parent.label == "Default orientation"
        assert_allclose(subset1.subset_state.center(), (1.75, 1.75))

        assert subset2.subset_state.xatt.parent.label == "Default orientation"
        assert_allclose(subset2.subset_state.roi.xmin, -0.25)
        assert_allclose(subset2.subset_state.roi.ymin, -0.25)
        assert_allclose(subset2.subset_state.roi.xmax, 1.75)
        assert_allclose(subset2.subset_state.roi.ymax, 1.75)


class TestDeleteWCSLayerWithSubset(BaseImviz_WCS_GWCS):
    """Regression test for https://jira.stsci.edu/browse/JDAT-3958"""

    def test_delete_wcs_layer_with_subset(self):
        lc_plugin = self.imviz.plugins['Orientation']
        lc_plugin.align_by = 'WCS'

        # Should automatically be applied as reference to first viewer.
        lc_plugin.set_north_up_east_left()

        # Create a rotated ellipse.
        reg = EllipsePixelRegion(
            PixCoord(3.5, 4.5), width=2, height=5, angle=Angle(30, 'deg')).to_sky(self.wcs_1)
        self.imviz.plugins['Subset Tools'].import_region(reg)

        # Switch back to Default Orientation.
        self.imviz.app._change_reference_data("Default orientation")

        # Delete N-up E-left reference data from GUI.
        self.imviz.app.data_item_remove("North-up, East-left")

        # Make sure rotated ellipse is still the same as before.
        out_reg_d = self.imviz.app.get_subsets(include_sky_region=True)['Subset 1'][0]['sky_region']
        assert_allclose(reg.center.ra.deg, out_reg_d.center.ra.deg)
        assert_allclose(reg.center.dec.deg, out_reg_d.center.dec.deg)
        assert_quantity_allclose(reg.height, out_reg_d.height)
        assert_quantity_allclose(reg.width, out_reg_d.width)
        assert_quantity_allclose(reg.angle, out_reg_d.angle)



import pytest
import numpy as np
from numpy.testing import assert_allclose
from regions import RectanglePixelRegion


def test_spectrum_at_spaxel_no_alt(cubeviz_helper, spectrum1d_cube_with_uncerts):
    cubeviz_helper.load_data(spectrum1d_cube_with_uncerts, data_label='test')

    flux_viewer = cubeviz_helper.app.get_viewer("flux-viewer")
    uncert_viewer = cubeviz_helper.app.get_viewer("uncert-viewer")
    spectrum_viewer = cubeviz_helper.app.get_viewer("spectrum-viewer")

    # Set the active tool to spectrumperspaxel
    flux_viewer.toolbar.active_tool = flux_viewer.toolbar.tools['jdaviz:spectrumperspaxel']
    x = 1
    y = 1
    assert len(flux_viewer.native_marks) == 2
    assert len(spectrum_viewer.data()) == 1

    assert_allclose(spectrum_viewer.get_limits(), (4.6228e-07, 4.6236e-07, 28, 92))

    # Move to spaxel location
    flux_viewer.toolbar.active_tool.on_mouse_move(
        {'event': 'mousemove', 'domain': {'x': x, 'y': y}, 'altKey': False})
    assert flux_viewer.toolbar.active_tool._mark in spectrum_viewer.figure.marks
    assert flux_viewer.toolbar.active_tool._mark.visible is True

    # Click on spaxel location
    flux_viewer.toolbar.active_tool.on_mouse_event(
        {'event': 'click', 'domain': {'x': x, 'y': y}, 'altKey': False})
    assert len(flux_viewer.native_marks) == 3
    assert len(spectrum_viewer.data()) == 2

    assert_allclose(spectrum_viewer.get_limits(),
                    (4.6228e-07, 4.6236e-07, 4, 15.6))  # Zoomed to spaxel
    spectrum_viewer.set_limits(
        x_min=4.623e-07, x_max=4.6232e-07, y_min=42, y_max=88)  # Zoom in X and Y

    # Check that a new subset was created
    subsets = cubeviz_helper.app.get_subsets()
    reg = subsets.get('Subset 1')[0]['region']
    assert len(subsets) == 1
    assert isinstance(reg, RectanglePixelRegion)

    # Move out of bounds
    flux_viewer.toolbar.active_tool.on_mouse_move(
        {'event': 'mousemove', 'domain': {'x': -1, 'y': -1}, 'altKey': False})
    assert flux_viewer.toolbar.active_tool._mark.visible is False

    # Mouse leave event
    flux_viewer.toolbar.active_tool.on_mouse_move(
        {'event': 'mouseleave', 'domain': {'x': x, 'y': y}, 'altKey': False})
    assert flux_viewer.toolbar.active_tool._mark.visible is False

    # Check that zoom in both X and Y is kept.
    assert_allclose(spectrum_viewer.get_limits(), (4.623e-07, 4.6232e-07, 42, 88))

    # Deselect tool
    flux_viewer.toolbar.active_tool = None
    assert len(flux_viewer.native_marks) == 3

    # Check in uncertainty viewer as well. Set mouseover here
    cubeviz_helper.app.session.application._tools['g-coords-info'].dataset.selected = 'none'
    uncert_viewer.toolbar.active_tool = uncert_viewer.toolbar.tools['jdaviz:spectrumperspaxel']
    uncert_viewer.toolbar.active_tool.on_mouse_move(
        {'event': 'mousemove', 'domain': {'x': x, 'y': y}, 'altKey': False})
    assert uncert_viewer.toolbar.active_tool._mark in spectrum_viewer.figure.marks
    assert uncert_viewer.toolbar.active_tool._mark.visible is True

    # Select specific data
    cubeviz_helper.app.session.application._tools['g-coords-info'].dataset.selected = 'test[FLUX]'
    uncert_viewer.toolbar.active_tool = uncert_viewer.toolbar.tools['jdaviz:spectrumperspaxel']
    uncert_viewer.toolbar.active_tool.on_mouse_move(
        {'event': 'mousemove', 'domain': {'x': x, 'y': y}, 'altKey': False})
    assert uncert_viewer.toolbar.active_tool._mark in spectrum_viewer.figure.marks
    assert uncert_viewer.toolbar.active_tool._mark.visible is True


@pytest.mark.parametrize("cube_type", ["Surface Brightness", "Flux"])
def test_spectrum_at_spaxel_altkey_true(cubeviz_helper, spectrum1d_cube,
                                        spectrum1d_cube_sb_unit, cube_type):

    # test is parameterize to test a cube that is in Jy / sr (Surface Brightness)
    # as well as Jy (Flux), to test that flux cubes, which are converted in the
    # parser to flux / pix^2 surface brightness cubes, both work correctly.

    if cube_type == 'Surface Brightness':
        cube = spectrum1d_cube_sb_unit
        cube_unit = 'Jy / sr'
    elif cube_type == 'Flux':
        cube = spectrum1d_cube
        cube_unit = 'Jy / pix2'

    cubeviz_helper.load_data(cube, data_label='test')

    flux_viewer = cubeviz_helper.app.get_viewer("flux-viewer")
    uncert_viewer = cubeviz_helper.app.get_viewer("uncert-viewer")
    spectrum_viewer = cubeviz_helper.app.get_viewer("spectrum-viewer")

    # Set the active tool to spectrumperspaxel
    flux_viewer.toolbar.active_tool = flux_viewer.toolbar.tools['jdaviz:spectrumperspaxel']

    assert len(flux_viewer.native_marks) == 2
    assert len(spectrum_viewer.data()) == 1

    # Check coordinate info panel
    sl = cubeviz_helper.plugins['Slice']
    sl.value = sl._obj.valid_indicator_values_sorted[1]
    assert flux_viewer.slice == 1
    label_mouseover = cubeviz_helper.app.session.application._tools['g-coords-info']
    label_mouseover._viewer_mouse_event(flux_viewer,
                                        {'event': 'mousemove', 'domain': {'x': 2, 'y': 1}})
    assert label_mouseover.as_text() == (f'Pixel x=02.0 y=01.0 Value +1.40000e+01 {cube_unit}',
                                         'World 13h39m59.9192s +27d00m00.7200s (ICRS)',
                                         '204.9996633015 27.0001999996 (deg)')

    # Click on spaxel location
    x = 1
    y = 1
    flux_viewer.toolbar.active_tool.on_mouse_event(
        {'event': 'click', 'domain': {'x': x, 'y': y}, 'altKey': False})
    assert len(flux_viewer.native_marks) == 3
    assert len(spectrum_viewer.data()) == 2

    # Check that subset was created
    subsets = cubeviz_helper.app.get_subsets()
    reg = subsets.get('Subset 1')[0]['region']
    assert len(subsets) == 1
    assert isinstance(reg, RectanglePixelRegion)

    # Using altKey should create a new subset
    x = 0
    y = 0
    flux_viewer.toolbar.active_tool.on_mouse_event(
        {'event': 'click', 'domain': {'x': x, 'y': y}, 'altKey': True})
    assert len(flux_viewer.native_marks) == 4
    assert len(spectrum_viewer.data()) == 3

    subsets = cubeviz_helper.app.get_subsets()
    reg2 = subsets.get('Subset 2')[0]['region']
    assert len(subsets) == 2
    assert isinstance(reg2, RectanglePixelRegion)

    # Make sure coordinate info panel did not change
    label_mouseover._viewer_mouse_event(flux_viewer,
                                        {'event': 'mousemove', 'domain': {'x': 1, 'y': 1}})
    assert label_mouseover.as_text() == (f'Pixel x=01.0 y=01.0 Value +1.30000e+01 {cube_unit}',
                                         'World 13h39m59.9461s +27d00m00.7200s (ICRS)',
                                         '204.9997755344 27.0001999998 (deg)')

    # Make sure linked pan mode works on all image viewers
    t_linkedpan = flux_viewer.toolbar.tools['jdaviz:pixelpanzoommatch']
    t_linkedpan.activate()
    # TODO: When Cubeviz uses Astrowidgets, can just use center_on() for this part.
    flux_viewer.set_limits(x_min=20, x_max=40, y_min=15, y_max=35)
    v = uncert_viewer
    assert v.state.zoom_center_x == flux_viewer.state.zoom_center_x
    assert v.state.zoom_center_y == flux_viewer.state.zoom_center_y
    assert v.state.zoom_radius == flux_viewer.state.zoom_radius
    t_linkedpan.deactivate()


def test_spectrum_at_spaxel_with_2d(cubeviz_helper):
    # Use cube with single slice
    x = np.array([[[1, 2, 3], [4, 5, 6]]])

    cubeviz_helper.load_data(x, data_label='test')

    flux_viewer = cubeviz_helper.app.get_viewer("flux-viewer")
    spectrum_viewer = cubeviz_helper.app.get_viewer("spectrum-viewer")

    # Set the active tool to spectrumperspaxel
    flux_viewer.toolbar.active_tool = flux_viewer.toolbar.tools['jdaviz:spectrumperspaxel']
    x = 0
    y = 0
    assert len(flux_viewer.native_marks) == 2
    assert len(spectrum_viewer.data()) == 1  # Spectrum (sum)

    # Click on spaxel location
    flux_viewer.toolbar.active_tool.on_mouse_event(
        {'event': 'click', 'domain': {'x': x, 'y': y}, 'altKey': False})
    assert len(flux_viewer.native_marks) == 3
    assert len(spectrum_viewer.data()) == 2  # Spectrum (sum), Spectrum (Subset 1, sum)

    # Deselect tool
    flux_viewer.toolbar.active_tool = None
    assert len(flux_viewer.native_marks) == 3



import pytest
import numpy as np
import astropy.units as u
from specutils import SpectralRegion


def test_spectralsubsetselect(specviz_helper, spectrum1d):
    # apply mask to spectrum to check selected subset is masked:
    mask = spectrum1d.flux < spectrum1d.flux.mean()
    spectrum1d.mask = mask

    specviz_helper.load_data(spectrum1d)
    sv = specviz_helper.app.get_viewer('spectrum-viewer')
    # create a "Subset 1" entry
    subset_plugin = specviz_helper.plugins['Subset Tools']
    subset_plugin.import_region(SpectralRegion(6500 * spectrum1d.spectral_axis.unit,
                                               7400 * spectrum1d.spectral_axis.unit))

    # model fitting uses the mixin
    p = specviz_helper.app.get_tray_item_from_name('g-model-fitting')
    assert len(p.spectral_subset.labels) == 2  # Entire Spectrum, Subset 1
    assert len(p.spectral_subset_items) == 2
    assert p.spectral_subset_selected == 'Entire Spectrum'
    assert p.spectral_subset_selected_has_subregions is False
    assert p.spectral_subset.selected_obj is None
    p.spectral_subset_selected = 'Subset 1'
    assert p.spectral_subset_selected_has_subregions is False
    assert p.spectral_subset.selected_obj is not None
    expected_min = spectrum1d.spectral_axis[spectrum1d.spectral_axis.value >= 6500][0]
    expected_max = spectrum1d.spectral_axis[spectrum1d.spectral_axis.value <= 7400][-1]
    np.testing.assert_allclose(expected_min.value, 6666.66666667, atol=1e-5)
    np.testing.assert_allclose(expected_max.value, 7333.33333333, atol=1e-5)
    assert p.spectral_subset.selected_min_max(spectrum1d) == (6500 * u.AA, 7400 * u.AA)

    # check selected subset mask available via API:
    expected_mask_with_spectral_subset = (
        (spectrum1d.wavelength.to_value(u.AA) < 6500) |
        (spectrum1d.wavelength.to_value(u.AA) > 7400)
    )
    assert np.all(
        expected_mask_with_spectral_subset == p.spectral_subset.selected_subset_mask
    )

    assert p.spectral_subset.app == p.app
    assert p.spectral_subset.spectrum_viewer == sv

    # line analysis uses custom components, one of which is still named spectral_subset
    p = specviz_helper.app.get_tray_item_from_name('specviz-line-analysis')
    assert len(p.spectral_subset.labels) == 2  # Entire Spectrum, Subset 1
    assert len(p.spectral_subset_items) == 2
    assert p.spectral_subset_selected == 'Entire Spectrum'
    assert p.spectral_subset.selected_obj is None
    p.spectral_subset_selected = 'Subset 1'
    assert p.spectral_subset.selected_obj is not None


@pytest.mark.filterwarnings('ignore:No observer defined on WCS')
def test_viewer_select(cubeviz_helper, spectrum1d_cube):
    app = cubeviz_helper.app
    app.add_data(spectrum1d_cube, 'test')
    app.add_data_to_viewer("flux-viewer", "test")
    fv = app.get_viewer("flux-viewer")
    sv = app.get_viewer("spectrum-viewer")

    # export plugin uses the mixin
    p = cubeviz_helper.plugins['Export']
    assert len(p.viewer.ids) == 3
    assert len(p.viewer.references) == 3
    assert len(p.viewer.labels) == 3
    assert p.viewer.selected_obj == fv

    # set by reference
    p.viewer = 'spectrum-viewer'
    assert p.viewer.selected_obj == sv

    # try setting based on id instead of reference
    p.viewer = p.viewer.ids[0]
    assert p.viewer.selected == p.viewer.labels[0]



from traitlets import Bool, Unicode
from glue.core.message import (DataCollectionAddMessage,
                               DataCollectionDeleteMessage)

from jdaviz.core.template_mixin import PluginTemplateMixin, AutoTextField
from jdaviz.core.user_api import ImporterUserApi

__all__ = ['BaseImporter', 'BaseImporterToDataCollection', 'BaseImporterToPlugin']


class BaseImporter(PluginTemplateMixin):
    def __init__(self, app, input, **kwargs):
        self._input = input
        super().__init__(app, **kwargs)

    @property
    def is_valid(self):
        # override by subclass
        return False  # pragma: nocover

    @property
    def input(self):
        return self._input

    @property
    def output(self):
        # override by subclass
        return self.input

    @property
    def target(self):
        raise NotImplementedError("Importer subclass must implement target")  # pragma: nocover

    def __call__(self):
        # override by subclass - should act on self.output and load into jdaviz
        raise NotImplementedError("Importer subclass must implement __call__")  # pragma: nocover

    @property
    def user_api(self):
        return ImporterUserApi(self)


class BaseImporterToDataCollection(BaseImporter):
    data_label_value = Unicode().tag(sync=True)
    data_label_default = Unicode().tag(sync=True)
    data_label_auto = Bool(True).tag(sync=True)
    data_label_invalid_msg = Unicode().tag(sync=True)

    def __init__(self, app, input, **kwargs):
        super().__init__(app, input, **kwargs)
        self.data_label_default = self._registry_label
        self.data_label = AutoTextField(self, 'data_label_value',
                                        'data_label_default',
                                        'data_label_auto',
                                        'data_label_invalid_msg')

        self.hub.subscribe(self, DataCollectionAddMessage,
                           handler=lambda _: self._on_label_changed())
        self.hub.subscribe(self, DataCollectionDeleteMessage,
                           handler=lambda _: self._on_label_changed())
        self.observe(self._on_label_changed, 'data_label_value')
        self._on_label_changed()

    @property
    def default_viewer(self):
        raise NotImplementedError("Importer subclass must implement default_viewer")  # noqa pragma: nocover

    @property
    def target(self):
        return self.default_viewer

    def _on_label_changed(self, msg={}):
        if not len(self.data_label_value.strip()):
            # strip will raise the same error for a label of all spaces
            self.data_label_invalid_msg = 'label must be provided'
            return

        # ensure the default label is unique for the data-collection
        self.data_label_default = self.app.return_unique_name(self.data_label_default)

        for data in self.app.data_collection:
            if self.data_label_value == data.label:
                self.data_label_invalid_msg = 'label already in use'
                return

        self.data_label_invalid_msg = ''

    def load_into_viewer(self, data_label):
        added = 0
        for viewer in self.app._jdaviz_helper.viewers.values():
            if data_label in viewer.data_menu.data_labels_unloaded:
                added += 1
                viewer.data_menu.add_data(data_label)
        if added == 0:
            print(f"*** will eventually create {self.default_viewer} and add data")

    def add_to_data_collection(self, data, data_label, show_in_viewer=True):
        if data_label is None:
            data_label = self.data_label_value
        self.app.add_data(data, data_label=data_label)
        if show_in_viewer:
            self.load_into_viewer(data_label)

    def __call__(self, data_label=None):
        self.add_to_data_collection(self.output, data_label, show_in_viewer=True)


class BaseImporterToPlugin(BaseImporter):
    @property
    def default_plugin(self):
        raise NotImplementedError("Importer subclass must implement default_plugin")  # noqa pragma: nocover

    @property
    def target(self):
        return self.default_plugin

    @property
    def has_default_plugin(self):
        return self.default_plugin in self.app._jdaviz_helper.plugins



from traitlets import Bool, Float, observe
from time import sleep

from jdaviz import Cubeviz
from jdaviz.app import Application
from jdaviz.core.config import get_configuration
from jdaviz.core.registries import tray_registry
from jdaviz.core.template_mixin import PluginTemplateMixin, skip_if_no_updates_since_last_active


@tray_registry('test-slow-plugin', label="Test Slow Plugin")
class SlowPlugin(PluginTemplateMixin):
    """
    empty plugin with a simulated slow method
    """
    template = ''
    uses_active_status = Bool(True).tag(sync=True)
    sleep_time = Float(0).tag(sync=True)

    def __init__(self, *args, **kwargs):
        # reset this manually in the test to see if slow_method ran since last manual reset
        self._method_ran = False
        super().__init__(*args, **kwargs)

    @observe('is_active', 'sleep_time')
    @skip_if_no_updates_since_last_active()
    def slow_method(self, *args):
        self._method_ran = True
        sleep(self.sleep_time)
        return


def test_is_active():
    # create an instance of an app with just our fake "Slow" Plugin
    config = get_configuration('cubeviz')
    config['tray'] = ['test-slow-plugin']

    cubeviz_app = Application(config)
    cubeviz_helper = Cubeviz(cubeviz_app)

    plg = cubeviz_helper.plugins['Test Slow Plugin']._obj
    plg.open_in_tray()
    assert plg.uses_active_status is True
    # there is no actual frontend here to ping (this would be True in a notebook)
    assert plg.is_active is False

    # so we need to fake the pings
    assert plg._ping_timestamp == 0
    assert plg.sleep_time == 0
    plg.vue_plugin_ping({})
    # since the observe on is_active/sleep_time returns immediately, the plugin is active
    assert plg.is_active is True
    assert plg._ping_timestamp > 0
    assert plg._method_ran is True

    # if there are no pings then the plugin will reset to inactive
    sleep(1)
    assert plg.is_active is False

    # if we set sleep_time while the plugin is inactive, the method should not fire
    plg._method_ran = False
    plg.sleep_time = 0.01
    assert plg._method_ran is False
    # but should immediately once it becomes active
    plg.vue_plugin_ping({})
    assert plg._method_ran is True



from collections.abc import Iterable
import itertools

from astropy import units as u
import numpy as np

from jdaviz.core.custom_units_and_equivs import (PIX2,
                                                 SPEC_PHOTON_FLUX_DENSITY_UNITS,
                                                 _eqv_pixar_sr,
                                                 _eqv_flux_to_sb_pixel,
                                                 _spectral_and_photon_flux_density_units)

__all__ = ["all_flux_unit_conversion_equivs", "check_if_unit_is_per_solid_angle",
           "combine_flux_and_angle_units", "convert_integrated_sb_unit",
           "create_equivalent_angle_units_list",
           "create_equivalent_flux_units_list",
           "create_equivalent_spectral_axis_units_list",
           "flux_conversion_general", "handle_squared_flux_unit_conversions",
           "supported_sq_angle_units", "spectral_axis_conversion",
           "units_to_strings"]


def all_flux_unit_conversion_equivs(pixar_sr=None, cube_wave=None):
    """
    Combines commonly used flux unit conversion equivalencies for
    translations between flux units and between flux and surface brightness
    units.

    - Flux to flux per square pixel
    - Flux to flux per steradian if ``pixar_sr`` is provided.
    - Spectral density conversions (e.g. Jy to erg/s/cm2/A), if ``cube_wave``
      is provided

    Parameters
    ----------
    pixar_sr : float, optional
        Pixel scale factor in steradians.
    cube_wave : `~astropy.units.Quantity`, optional
        A reference wavelength or frequency value(s).

    Returns
    -------
    equivs : list
        List of equivalencies.
    """

    equivs = _eqv_flux_to_sb_pixel()

    if pixar_sr is not None:
        equivs += _eqv_pixar_sr(pixar_sr)

    if cube_wave is not None:
        equivs += u.spectral_density(cube_wave)

    return equivs


def viewer_flux_conversion_equivalencies(values, spec):
    """
    Generate a list of flux and surface brightness unit conversion equivalencies
    specifically for converting units in the viewer, accounting for the special
    case of viewer limits that may need to index the spectral axis or list of
    pixel scale factors.

    This function assumes that if exactly two values are being converted, they
    represent the y-axis limits of the viewer. In this case, the first spectral
    axis value is used for spectral density conversions, and the min/max value
    of pixel scale factor are used if there are more than 2 values in
    meta['_pixel_scale_factor'].

    Parameters:
    ----------
    values : array-like
        The values to be converted, which may represent flux or surface brightness.
    spec : Spectrum1D

    Returns:
    -------
    equivs : list
        A list of unit equivalencies for flux and surface brightness conversions.
    """

    # if we are converting only 2 values, assume it is a viewer y limits case.
    is_viewer_limits = len(values) == 2

    # for viewer limits case, use only the 0th spectral axis value for spectral_density
    spectral_values = spec.spectral_axis
    if not np.isscalar(values) and is_viewer_limits:
        spectral_values = spectral_values[0]

    # Need this for setting the y-limits but values from viewer might be downscaled
    if len(values) != spectral_values.size:
        spectral_values = spec.spectral_axis[0]

    # Next, pixel scale factor
    pix_fac = None
    if '_pixel_scale_factor' in spec.meta:
        pix_fac = spec.meta['_pixel_scale_factor']

        if isinstance(pix_fac, u.Quantity):
            pix_fac = pix_fac.value

        # If 2 values are being converted (considered to be viewer y limits),
        # use min and max scale factors.
        if is_viewer_limits and isinstance(pix_fac, Iterable):
            pix_fac = [min(pix_fac), max(pix_fac)]

    # combine scale factor and spectral axis for u.spectral with other flux<>sb equivalencies
    equivs = all_flux_unit_conversion_equivs(pix_fac, spectral_values)

    return equivs


def check_if_unit_is_per_solid_angle(unit, return_unit=False):
    """
    Check if a given Unit or unit string (that can be converted to
    a Unit) represents some unit per solid angle. If 'return_unit'
    is True, then a Unit of the solid angle will be returned (or
    None if no solid angle is present in the denominator).

    Parameters
    ----------
    unit : str or u.Unit
        u.Unit object or string representation of unit.
    return_unit : bool
        If True, the u.Unit of the solid angle unit will
        be returned (or None if unit is not a solid angle).

    Returns
    -------
    result : `~astropy.units.Unit`, bool, or `None`
        See explanation in ``return_unit``.

    Raises
    ------
    ValueError
        Invalid input.

    Examples
    --------
    >>> check_if_unit_is_per_solid_angle('erg / (s cm^2 sr)')
    True
    >>> check_if_unit_is_per_solid_angle('erg / s cm^2')
    False
    >>> check_if_unit_is_per_solid_angle('Jy * sr^-1')
    True

    """

    # first, convert string to u.Unit obj.
    # this will take care of some formatting consistency like
    # turning something like Jy / (degree*degree) to Jy / deg**2
    # and erg sr^1 to erg / sr
    if isinstance(unit, (u.core.Unit, u.core.CompositeUnit,
                         u.core.IrreducibleUnit)):
        unit_str = unit.to_string()
    elif isinstance(unit, str):
        # convert string>unit>string to remove any formatting inconsistencies
        unit = u.Unit(unit)
        unit_str = unit.to_string()
    else:
        raise ValueError('Unit must be u.Unit, or string that can be converted into a u.Unit')

    if '/' in unit_str:
        # input unit might be comprised of several units in denom. so check all.
        denom = unit_str.split('/')[-1].split()

        # find all combos of one or two units, to catch cases where there are
        # two different units of angle in the denom that might comprise a solid
        # angle when multiplied.
        for i in [combo for length in (1, 2) for combo in itertools.combinations(denom, length)]:
            # turn tuple of 1 or 2 units into a string, and turn that into a u.Unit
            # to check type
            new_unit_str = ' '.join(i).translate(str.maketrans('', '', '()'))
            new_unit = u.Unit(new_unit_str)
            if new_unit.physical_type == 'solid angle' or new_unit == PIX2 or new_unit_str == 'spaxel':  # noqa
                # square pixel and spaxel should be considered square angle units
                if return_unit:  # area units present and requested to be returned
                    return new_unit
                return True  # area units present but not requested to be returned

    # in the case there are no area units, but return units were requested
    if return_unit:
        return None

    # and if there are no area units, and return units were NOT requested.
    return False


def combine_flux_and_angle_units(flux_units, angle_units):
    """
    Combine (list of) flux_units and angle_units to create a list of string
    representations of surface brightness units. The returned strings will be in
    the same format as the astropy unit to_string() of the unit, for consistency.
    """
    if not isinstance(flux_units, list):
        flux_units = [flux_units]
    if not isinstance(angle_units, list):
        angle_units = [angle_units]

    return [(u.Unit(flux) / u.Unit(angle)).to_string() for flux in flux_units
            for angle in angle_units]


def create_equivalent_angle_units_list(solid_angle_unit):

    """
    Return valid angles that ``solid_angle_unit`` (which should be a solid angle
    physical type, or square pixel), can be translated to in the unit conversion
    plugin. These options will populate the dropdown menu for 'angle unit' in
    the  Unit Conversion plugin.

    Parameters
    ----------
    solid_angle_unit : str or u.Unit
        Unit object or string representation of unit that is a ``solid angle``
        or square pixel physical type.

    Returns
    -------
    equivalent_angle_units : list of str
        String representation of units that ``solid_angle_unit`` can be
        translated to.

    """

    if solid_angle_unit is None or solid_angle_unit is PIX2:
        # if there was no solid angle in the unit when calling this function
        # can only represent that unit as per square pixel
        return ['pix^2']

    # cast to unit then back to string to account for formatting inconsistencies
    # in strings that represent units
    if isinstance(solid_angle_unit, str):
        solid_angle_unit = u.Unit(solid_angle_unit)
    unit_str = solid_angle_unit.to_string()

    # uncomment and expand this list once translating between solid
    # angles and between solid angle and solid pixel is enabled
    # equivalent_angle_units = ['sr', 'pix^2']
    equivalent_angle_units = []
    if unit_str not in equivalent_angle_units:
        equivalent_angle_units += [unit_str]

    return equivalent_angle_units


def create_equivalent_flux_units_list(flux_unit):
    """
    Get all possible conversions for flux from flux_unit, to populate 'flux'
    dropdown menu in the unit conversion plugin.

    If flux_unit is a spectral or photon density (i.e., convertable to units in
    SPEC_PHOTON_FLUX_DENSITY_UNITS), then the loaded unit and all of the
    units in SPEC_PHOTON_FLUX_DENSITY_UNITS.

    If the loaded flux unit is count, dimensionless_unscaled, DN, e/s, then
    there will be no additional items available for unit conversion and the
    only item in the dropdown will be the native unit.
    """

    flux_unit_str = flux_unit.to_string()

    # if flux_unit is a spectral or photon flux density unit, then the flux unit
    # dropdown options should be the loaded unit (which may have a different
    # prefix e.g nJy) in addition to items in SPEC_PHOTON_FLUX_DENSITY_UNITS
    equiv = u.spectral_density(1 * u.m)  # unit doesn't matter, not evaluating
    for un in SPEC_PHOTON_FLUX_DENSITY_UNITS:
        if flux_unit.is_equivalent(un, equiv):
            if flux_unit_str not in SPEC_PHOTON_FLUX_DENSITY_UNITS:
                return SPEC_PHOTON_FLUX_DENSITY_UNITS + [flux_unit_str]
            else:
                return SPEC_PHOTON_FLUX_DENSITY_UNITS

    else:
        # for any other units, including counts, DN, e/s, DN /s, etc,
        # no other conversions between flux units available as we only support
        # conversions to and from spectral and photon flux density flux unit.
        # dropdown will only contain one item (the input unit)
        return [flux_unit_str]


def create_equivalent_spectral_axis_units_list(spectral_axis_unit,
                                               exclude=[u.jupiterRad, u.earthRad,
                                                        u.solRad, u.lyr, u.AU,
                                                        u.pc, u.Bq, u.micron,
                                                        u.lsec]):
    """Get all possible conversions from current spectral_axis_unit."""
    if spectral_axis_unit in (u.pix, u.dimensionless_unscaled):
        return [spectral_axis_unit.to_string()]

    # Get unit equivalencies.
    try:
        curr_spectral_axis_unit_equivalencies = spectral_axis_unit.find_equivalent_units(
            equivalencies=u.spectral())
    except u.core.UnitConversionError:
        return []

    # Get local units.
    locally_defined_spectral_axis_units = ['Angstrom', 'nm',
                                           'um', 'Hz', 'erg']
    local_units = [u.Unit(unit) for unit in locally_defined_spectral_axis_units]

    # Remove overlap units.
    curr_spectral_axis_unit_equivalencies = list(set(curr_spectral_axis_unit_equivalencies)
                                                 - set(local_units + exclude))

    # Convert equivalencies into readable versions of the units and sorted alphabetically.
    spectral_axis_unit_equivalencies_titles = sorted(units_to_strings(
        curr_spectral_axis_unit_equivalencies))

    # Concatenate both lists with the local units coming first.
    return sorted(units_to_strings(local_units)) + spectral_axis_unit_equivalencies_titles


def flux_conversion_general(values, original_unit, target_unit,
                            equivalencies=None, with_unit=True):
    """
    Converts ``values`` from ``original_unit`` to ``target_unit`` using the
    provided ``equivalencies`` while handling special cases where direct unit
    conversion is not possible. This function is designed to account for
    scenarios like conversions involving flux to surface brightness that
    also require a ``u.spectral_density`` equivalency, conversions between
    per-square pixel surface brightnesses that don't convert directly, and other
    flux to surface brightness conversions.

    This function should be used for unit conversions when possible instead of
    directly using Astropy's ``unit.to()``, as it handles additional logic for
    special cases.

    Parameters
    ----------
    values : array-like or float
        The numerical values to be converted.
    original_unit : `~astropy.units.Unit` or str
        The unit of the input values.
    target_unit : `~astropy.units.Unit` or str
        The desired unit to convert to.
    equivalencies : list of equivalencies, optional
        Unit equivalencies to apply during the conversion.
    with_unit : bool, optional
        If True, the returned value retains its unit. If False, only the
        numerical values are returned.

    Returns
    -------
    converted_values : `~astropy.units.Quantity` or float
        The converted values, with or without units based on ``with_unit``.

    Raises
    ------
    astropy.units.UnitConversionError
        If the conversion between ``original_unit`` and ``target_unit`` fails
        despite the provided equivalencies.

    """

    # we set surface brightness choices and selection before flux, which can
    # cause a dimensionless translation attempt at instantiation
    if not target_unit:
        return values

    if original_unit == target_unit:
        if not with_unit:
            return values
        return values * original_unit

    if isinstance(original_unit, str):
        original_unit = u.Unit(original_unit)
    if isinstance(target_unit, str):
        target_unit = u.Unit(target_unit)

    solid_angle_in_orig = check_if_unit_is_per_solid_angle(original_unit,
                                                           return_unit=True)
    solid_angle_in_targ = check_if_unit_is_per_solid_angle(target_unit,
                                                           return_unit=True)

    with u.set_enabled_equivalencies(equivalencies):

        # first possible case we want to catch before trying to translate: both
        # the original and target unit are per-pixel-squared SB units
        # and also require an additional equivalency, so we need to multiply out
        # the pix2 before conversion and re-apply. if this doesn't work, something else
        # is going on (missing equivalency, etc)
        if solid_angle_in_orig == solid_angle_in_targ == PIX2:
            converted_values = (values * (original_unit * PIX2)).to(target_unit * PIX2)
            converted_values = converted_values / PIX2  # re-apply pix2 unit
        else:
            try:
                # if units can be converted straight away with provided
                # equivalencies, return converted values
                converted_values = (values * original_unit).to(target_unit)
            except u.UnitConversionError:
                # the only other case where units with the correct equivs wouldn't
                # convert directly is if one unit is a flux and one is a sb and
                # they also require an additional equivalency
                if not bool(solid_angle_in_targ) == bool(solid_angle_in_orig):
                    converted_values = (values * original_unit * (solid_angle_in_orig or 1)).to(target_unit * (solid_angle_in_targ or 1))  # noqa
                    converted_values = (converted_values / (solid_angle_in_orig or 1)).to(target_unit)  # noqa
                else:
                    raise u.UnitConversionError(f'Could not convert {original_unit} to {target_unit} with provided equivalencies.')  # noqa

        if not with_unit:
            return converted_values.value

        return converted_values


def handle_squared_flux_unit_conversions(value, original_unit=None,
                                         target_unit=None, equivalencies=None):
    """
    Handles conversions between squared flux or surface brightness units
    that cannot be directly converted, even with the correct equivalencies.

    This function is specifically designed to address cases where squared
    units, such as (MJy/sr)**2 to (Jy/sr)**2, appear in contexts like
    variance columns of aperture photometry output tables. When additional
    equivalencies are required, direct conversion may fail, so this workaround.
    is required.

    Parameters
    ----------
    value : array or float
        The numerical values to be converted.
    original_unit : `astropy.units.Unit` or str
        The unit of the input values before conversion.
    target_unit : `astropy.units.Unit` or str
        The desired unit for the converted values.
    equivalencies : list of equivalencies
        Unit equivalencies to apply during the conversion.

    Returns
    -------
    converted : `~astropy.units.Quantity`
        The converted values, expressed in the ``target_unit``.
    """

    # get scale factor between non-squared units
    converted = flux_conversion_general(1.,
                                        original_unit ** 0.5,
                                        target_unit ** 0.5,
                                        equivalencies,
                                        with_unit=False)

    # square conversion factor and re-apply squared unit
    converted = converted ** 2 * value * target_unit

    return converted


def spectral_axis_conversion(values, original_units, target_units):
    eqv = u.spectral() + u.pixel_scale(1*u.pix)
    return (values * u.Unit(original_units)).to_value(u.Unit(target_units), equivalencies=eqv)


def supported_sq_angle_units(as_strings=False):
    """
    Returns a list of squared angle units supported by the app. If a new
    solid angle is added into unit conversion logic (e.g., square degree), it
    should be added here.
    """

    units = [PIX2, u.sr]
    if as_strings:
        units = units_to_strings(units)
    return units


def units_to_strings(unit_list):
    """Convert equivalencies into readable versions of the units.

    Parameters
    ----------
    unit_list : list
        List of either `astropy.units.Unit` or strings that can be converted
        to `astropy.units.Unit`.

    Returns
    -------
    result : list
        A list of the units with their best (i.e., most readable) string version.
    """
    return [u.Unit(unit).to_string() for unit in unit_list]


def convert_integrated_sb_unit(u1, spectral_axis_unit, desired_freq_unit, desired_length_unit):
    """
    Converts an integrated surface brightness unit (moment 0 unit) to a surface
    brighntess unit that is compatible with the spectral axis unit that the surface
    brightness was integrated over.

    This function adjusts an integrated flux unit to ensure compatibility with a given
    spectral axis unit (e.g., frequency or wavelength). The function handles conversions
    based on the physical type of the flux unit (per-frequency or per-wavelength) and
    the provided spectral axis unit.

    Parameters
    ----------
    u1 : astropy.units.Unit
        The unit of the integrated flux that needs conversion.

    spectral_axis_unit : astropy.units.Unit
        The unit of the spectral axis over which the flux was integrated (e.g., Angstrom
        for wavelength or Hz for frequency).

    Returns
    -------
    astropy.units.Unit
        The converted flux unit compatible with the given spectral axis unit. If the
        units are already compatible, the input unit ``u1`` is returned unchanged.
    """

    uu = u1 / spectral_axis_unit

    # multiply solid angle unit out of surface brightness to compare just flux components
    flux = uu * check_if_unit_is_per_solid_angle(uu.unit, return_unit=True)

    # then check if flux unit is a per-frequency or per-wavelength flux unit
    wav_units = _spectral_and_photon_flux_density_units(wav_only=True, as_units=True)
    freq_units = _spectral_and_photon_flux_density_units(freq_only=True, as_units=True)
    if np.any([flux.unit.is_equivalent(x) for x in wav_units]):
        flux_unit_type = 'length'
    elif np.any([flux.unit.is_equivalent(x) for x in freq_units]):
        flux_unit_type = 'frequency'

    if (spectral_axis_unit.physical_type != flux_unit_type):
        if flux_unit_type == 'length':
            spec_axis_conversion_scale_factor = (1*spectral_axis_unit).to(desired_length_unit,
                                                                          u.spectral())
        elif flux_unit_type == 'frequency':
            spec_axis_conversion_scale_factor = (1*spectral_axis_unit).to(desired_freq_unit,
                                                                          u.spectral())
    else:
        return u1  # units are compatible, return input

    return uu * spec_axis_conversion_scale_factor



from itertools import chain
import os
from traitlets import Any, Dict, Bool, List, Unicode, Float, observe

import numpy as np
from glue_jupyter.common.toolbar_vuetify import read_icon
from echo import delay_callback
from matplotlib.colors import hex2color

from jdaviz.core.registries import tray_registry
from jdaviz.core.template_mixin import (
    PluginTemplateMixin, LayerSelect, ViewerSelectMixin
)
from jdaviz.core.tools import ICON_DIR
from jdaviz.core.user_api import PluginUserApi
from jdaviz.configs.default.plugins.data_quality.dq_utils import (
    decode_flags, generate_listed_colormap, dq_flag_map_paths, load_flag_map
)


__all__ = ['DataQuality']

telescope_names = {
    "jwst": "JWST",
    "roman": "Roman",
    "hst-stis": "HST/STIS",
    "hst-acs": "HST/ACS",
    "hst-wfc3-uvis": "HST/WFC3-UVIS",
    "hst-wfc3-ir": "HST/WFC3-IR",
    "hst-cos": "HST/COS",
}


@tray_registry('g-data-quality', label="Data Quality", viewer_requirements="image")
class DataQuality(PluginTemplateMixin, ViewerSelectMixin):
    """
    See the :ref:`Data Quality Plugin Documentation <imviz-data-quality>` for more details.

    Only the following attributes and methods are available through the
    :ref:`public plugin API <plugin-apis>`:

    * :meth:`~jdaviz.core.template_mixin.PluginTemplateMixin.show`
    * :meth:`~jdaviz.core.template_mixin.PluginTemplateMixin.open_in_tray`
    * :meth:`~jdaviz.core.template_mixin.PluginTemplateMixin.close_in_tray`
    * ``science_layer`` (:class:`~jdaviz.core.template_mixin.LayerSelect`)
    * ``dq_layer`` (:class:`~jdaviz.core.template_mixin.LayerSelect`):
        data quality layer corresponding to the science data in ``science_layer``
    * ``dq_layer_opacity``: Opacity of the data quality layer.
    * ``decoded_flags``: List of decoded flags from the selected flag map.
    * ``flags_filter``: List of flags to display.
    * ``flag_map``: Name of the selected flag map.
    * ``flag_map_definitions_selected``: Dictionary of the selected flag map.
    """
    template_file = __file__, "data_quality.vue"

    # `layer` is the science data layer
    science_layer_multiselect = Bool(False).tag(sync=True)
    science_layer_items = List().tag(sync=True)
    science_layer_selected = Any().tag(sync=True)  # Any needed for multiselect

    # `dq_layer` is the data quality layer corresponding to the
    # science data in `science_layer`
    dq_layer_multiselect = Bool(False).tag(sync=True)
    dq_layer_items = List().tag(sync=True)
    dq_layer_selected = Any().tag(sync=True)  # Any needed for multiselect
    dq_layer_opacity = Float(0.9).tag(sync=True)  # Any needed for multiselect

    flag_map_definitions = Dict().tag(sync=True)
    flag_map_selected = Any().tag(sync=True)
    flag_map_definitions_selected = Dict().tag(sync=True)
    flag_map_items = List().tag(sync=True)
    decoded_flags = List().tag(sync=True)
    flags_filter = List().tag(sync=True)

    icons = Dict().tag(sync=True)
    icon_radialtocheck = Unicode(read_icon(os.path.join(ICON_DIR, 'radialtocheck.svg'), 'svg+xml')).tag(sync=True)  # noqa
    icon_checktoradial = Unicode(read_icon(os.path.join(ICON_DIR, 'checktoradial.svg'), 'svg+xml')).tag(sync=True)  # noqa

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # description displayed under plugin title in tray
        self._plugin_description = 'Data Quality layer visualization options.'

        self.icons = {k: v for k, v in self.app.state.icons.items()}

        self.science_layer = LayerSelect(
            self, 'science_layer_items', 'science_layer_selected',
            'viewer_selected', 'science_layer_multiselect',
            is_root=True, has_children=True
        )

        self.dq_layer = LayerSelect(
            self, 'dq_layer_items', 'dq_layer_selected',
            'viewer_selected', 'dq_layer_multiselect', is_root=False,
            is_child_of=self.science_layer.selected
        )

        self.load_default_flag_maps()
        self.init_decoding()
        self._set_irrelevant()
        self._update_available_viewers()

    def _update_available_viewers(self):
        if not hasattr(self, 'viewer'):
            return

        def is_image_viewer(viewer):
            from jdaviz.configs.imviz.plugins.viewers import ImvizImageView
            from jdaviz.configs.cubeviz.plugins.viewers import CubevizImageView
            from jdaviz.configs.rampviz.plugins.viewers import RampvizImageView

            return isinstance(viewer, (ImvizImageView, CubevizImageView, RampvizImageView))

        viewer_filter_names = [filt.__name__ for filt in self.viewer.filters]
        if 'is_image_viewer' not in viewer_filter_names:
            self.viewer.add_filter(is_image_viewer)

        # by default, select all image viewers to sync each DQ layer's
        # options across all viewers that they're visible in:
        self.viewer_multiselect = True
        self.viewer.select_all()

    @observe('dq_layer_items')
    def _set_irrelevant(self, *args):
        self._update_available_viewers()

        children_available = any([
            len(assoc['children']) > 0
            for label, assoc in getattr(self.app, '_data_associations', {}).items()
        ])
        self.irrelevant_msg = (
            '' if children_available else
            "No Data Quality layers available."
        )

    @observe('science_layer_selected')
    def update_dq_layer(self, *args):
        if not hasattr(self, 'dq_layer'):
            return

        self.dq_layer.filter_is_child_of = self.science_layer_selected
        self.dq_layer._update_items()

        # listen for changes on the image opacity, and update the
        # data quality layer opacity on changes to the science layer opacity
        plot_options = self.app.get_tray_item_from_name('g-plot-options')
        plot_options.observe(self.update_opacity, 'image_opacity_value')

    def load_default_flag_maps(self):
        for name in dq_flag_map_paths:
            self.flag_map_definitions[name] = load_flag_map(name)
            self.flag_map_items = self.flag_map_items + [telescope_names[name]]

    @property
    def dq_layer_selected_flattened(self):
        if not hasattr(self, 'dq_layer'):
            return []

        selected_dq = self.dq_layer.selected_obj
        if not len(selected_dq):
            return []

        if isinstance(selected_dq, list) and isinstance(selected_dq[0], list):
            # flatten a nested list:
            selected_dq = list(chain.from_iterable(selected_dq))
        elif not isinstance(selected_dq, list):
            # if it's a single layer, make it a list:
            selected_dq = [selected_dq]
        return selected_dq

    @property
    def unique_flags(self):
        selected_dq = self.dq_layer_selected_flattened
        if selected_dq is None or not len(selected_dq):
            return []

        dq = selected_dq[0].get_image_data()

        return np.unique(dq[~np.isnan(dq)])

    @property
    def validate_flag_decode_possible(self):
        return (
            self.flag_map_selected is not None and
            len(self.dq_layer.selected_obj) > 0 and
            len(self.unique_flags) > 0
        )

    @observe('flag_map_selected')
    def update_flag_map_definitions_selected(self, event):
        flag_map_key = self.flag_map_selected.lower().replace('/', '-')
        selected = self.flag_map_definitions[flag_map_key]
        self.flag_map_definitions_selected = selected

        # clear decoded_flags with a meaningless one:
        self.init_decoding()
        self._update_cmap()

    @observe('dq_layer_selected')
    def init_decoding(self, event={}, viewers=None):
        if not self.validate_flag_decode_possible:
            return

        unique_flags = self.unique_flags
        cmap, rgba_colors = generate_listed_colormap(n_flags=len(unique_flags))
        self.decoded_flags = decode_flags(
            flag_map=self.flag_map_definitions_selected,
            unique_flags=unique_flags,
            rgba_colors=rgba_colors
        )
        self.send_state('decoded_flags')
        dq_layers = self.get_dq_layers(viewers=viewers)

        for dq_layer in dq_layers:
            dq_layer.composite._allow_bad_alpha = True

            # for cubeviz, also change uncert-viewer defaults to
            # map the out-of-bounds regions to the cmap's `bad` color:
            if self.app.config in ('cubeviz', 'rampviz'):
                viewer = self.app.get_viewer(
                    getattr(
                        self.app._jdaviz_helper,
                        '_default_uncert_viewer_reference_name', 'level-2'
                    )
                )

                for layer in viewer.layers:
                    # allow bad alpha for image layers, not subsets:
                    if not hasattr(layer, 'subset_array'):
                        layer.composite._allow_bad_alpha = True
                        layer.force_update()

            flag_bits = np.array([flag['flag'] for flag in self.decoded_flags])

            dq_layer.state.stretch = 'lookup'
            stretch_object = dq_layer.state.stretch_object
            stretch_object.flags = flag_bits

            with delay_callback(dq_layer.state, 'alpha', 'cmap', 'v_min', 'v_max'):
                if len(flag_bits):
                    dq_layer.state.v_min = min(flag_bits)
                    dq_layer.state.v_max = max(flag_bits)

                dq_layer.state.alpha = self.dq_layer_opacity
                dq_layer.state.cmap = cmap

    def get_dq_layers(self, viewers=None):
        if self.dq_layer_selected == '':
            return

        if viewers is None:
            viewers = self.viewer.selected_obj

        if not hasattr(viewers, '__len__'):
            viewers = [viewers]

        dq_layers = [
            layer for viewer in viewers
            for layer in viewer.layers if
            layer.layer.label == self.dq_layer_selected
        ]
        return dq_layers

    def get_science_layers(self, viewers=None):

        if viewers is None:
            viewers = self.viewer.selected_obj

        if not hasattr(viewers, '__len__'):
            viewers = [viewers]

        science_layers = [
            layer for viewer in viewers
            for layer in viewer.layers if
            layer.layer.label == self.science_layer_selected
        ]
        return science_layers

    @observe('dq_layer_opacity')
    def update_opacity(self, event={}):
        science_layers = self.get_science_layers()
        selected_dq = self.dq_layer_selected_flattened

        if len(selected_dq):
            for sci_layer, dq_layer in zip(science_layers, selected_dq):
                # DQ opacity is a fraction of the science layer's opacity:
                dq_layer.state.alpha = self.dq_layer_opacity * sci_layer.state.alpha

    @observe('decoded_flags', 'flags_filter')
    def _update_cmap(self, event={}, viewers=None):
        dq_layers = self.get_dq_layers(viewers=viewers)

        if dq_layers is None:
            return

        flag_bits = np.array([flag['flag'] for flag in self.decoded_flags])
        rgb_colors = [hex2color(flag['color']) for flag in self.decoded_flags]

        hidden_flags = np.array([
            flag['flag'] for flag in self.decoded_flags

            # hide the flag if the visibility toggle is False:
            if not flag['show'] or

            # hide the flag if `flags_filter` has entries but not this one:
            (
                len(self.flags_filter) and
                not np.isin(
                    list(map(int, flag['decomposed'].keys())),
                    list(self.flags_filter)
                ).any()
            )
        ])

        for dq_layer in dq_layers:
            with delay_callback(
                    dq_layer.state, 'v_min', 'v_max', 'alpha', 'stretch', 'cmap'
            ):
                # set correct stretch and limits:
                # dq_layer.state.stretch = 'lookup'
                stretch_object = dq_layer.state.stretch_object
                stretch_object.flags = flag_bits
                stretch_object.dq_array = dq_layer.get_image_data()
                stretch_object.hidden_flags = hidden_flags

                # update the colors of the listed colormap without
                # reassigning the layer.state.cmap object
                cmap = dq_layer.state.cmap
                cmap.colors = rgb_colors
                cmap._init()

                # trigger updates to cmap in viewer:
                dq_layer.update()

                if len(flag_bits):
                    dq_layer.state.v_min = min(flag_bits)
                    dq_layer.state.v_max = max(flag_bits)

                dq_layer.state.alpha = self.dq_layer_opacity

    def update_visibility(self, index):
        self.decoded_flags[index]['show'] = not self.decoded_flags[index]['show']
        self.vue_update_cmap()

    def vue_update_cmap(self):
        self.send_state('decoded_flags')
        self._update_cmap()

    def vue_update_visibility(self, index):
        self.update_visibility(index)

    def update_color(self, index, color):
        self.decoded_flags[index]['color'] = color
        self.vue_update_cmap()

    def vue_update_color(self, args):
        index, color = args
        self.update_color(index, color)

    @observe('science_layer_selected')
    def mission_or_instrument_from_meta(self, event):
        if not hasattr(self, 'science_layer'):
            return

        layer = self.get_science_layers()
        if not len(layer):
            return

        # this is defined for JWST and ROMAN, should be upper case:
        telescope = layer[0].layer.meta.get('telescope', None)

        if telescope is None:
            # for spectral cubes in Cubeviz:
            telescope = layer[0].layer.meta.get('_primary_header', {}).get('TELESCOP', None)

        if telescope is not None:
            primary_header = layer[0].layer.meta.get('_primary_header', {})
            if telescope == 'HST':
                instrument = primary_header.get('INSTRUME', None)
                if instrument == 'WFC3':
                    detector = primary_header.get('DETECTOR', None)
                else:
                    detector = None

                telescope = '-'.join(
                    i for i in [telescope, instrument, detector]
                    if i is not None
                )

            flag_map_to_select = telescope_names.get(telescope.lower())
            self.flag_map_selected = flag_map_to_select

    def vue_hide_all_flags(self, event):
        for flag in self.decoded_flags:
            flag['show'] = False

        self.vue_update_cmap()

    def vue_clear_flags_filter(self, event):
        self.flags_filter = []
        self.vue_update_cmap()

    def vue_show_all_flags(self, event):
        for flag in self.decoded_flags:
            flag['show'] = True

        self.flags_filter = []
        self.vue_update_cmap()

    @property
    def user_api(self):
        return PluginUserApi(
            self,
            expose=(
                'science_layer', 'dq_layer',
                'decoded_flags', 'flags_filter',
                'dq_layer_opacity',
                'flag_map_definitions_selected',
            )
        )



import numpy as np
import pytest
from astropy import units as u
from astropy.wcs import WCS
from numpy.testing import assert_allclose
from regions import PixCoord, CirclePixelRegion, RectanglePixelRegion
from specutils import Spectrum1D

from jdaviz.core.custom_units_and_equivs import PIX2, SPEC_PHOTON_FLUX_DENSITY_UNITS


def cubeviz_wcs_dict():
    # returns a WCS obj and dictionary used for cubeviz tests
    wcs_dict = {"CTYPE1": "WAVE-LOG", "CTYPE2": "DEC--TAN", "CTYPE3": "RA---TAN",
                "CRVAL1": 4.622e-7, "CRVAL2": 27, "CRVAL3": 205,
                "CDELT1": 8e-11, "CDELT2": 0.0001, "CDELT3": -0.0001,
                "CRPIX1": 0, "CRPIX2": 0, "CRPIX3": 0, "PIXAR_SR": 8e-11}
    w = WCS(wcs_dict)
    return w, wcs_dict


@pytest.mark.parametrize("angle_unit", [u.sr, PIX2])
def test_basic_unit_conversions(cubeviz_helper, angle_unit):
    """
    Basic test for changing flux units for a cube loaded in Jy to
    all available flux units. Checks that the the conversion does
    not produce any tracebacks. Tests conversions between all units
    in :
    ['Jy', 'mJy', 'uJy', 'MJy', 'W / (Hz m2)', 'eV / (Hz s m2)',
     'erg / (Hz s cm2)', 'erg / (Angstrom s cm2)',
     'ph / (Angstrom s cm2)', 'ph / (Hz s cm2)']

    Parametrized over both available solid angle units (pix2 and sr).
    """

    # load cube with flux units of MJy
    w, wcs_dict = cubeviz_wcs_dict()
    flux = np.ones((3, 4, 5), dtype=np.float32)
    cube = Spectrum1D(flux=flux * (u.MJy / angle_unit), wcs=w, meta=wcs_dict)
    cubeviz_helper.load_data(cube, data_label="test")
    viewer = cubeviz_helper.app.get_viewer("spectrum-viewer")

    # get all available flux units for translation. Since cube is loaded
    # in Jy, this will be all items in 'spectral_and_photon_flux_density_units'

    uc_plg = cubeviz_helper.plugins['Unit Conversion']
    ap_plg = cubeviz_helper.plugins["Aperture Photometry"]._obj
    label_mouseover = cubeviz_helper.app.session.application._tools['g-coords-info']

    cubeviz_helper.plugins['Subset Tools'].import_region(RectanglePixelRegion(PixCoord(1, 1), 1, 1))
    ap_plg.background_selected = "Subset 1"

    for flux_unit in SPEC_PHOTON_FLUX_DENSITY_UNITS:
        if flux_unit == 'MJy':
            if angle_unit == u.sr:
                ans = 9.6e-10  # 12 * 8e-11
            else:
                ans = 12  # 12 * 1
            bg_ans = 1  # MJy / angle_unit
        elif flux_unit == 'Jy':
            if angle_unit == u.sr:
                ans = 9.6e-04
            else:
                ans = 1.2e+07
            bg_ans = 1e6
        elif flux_unit == 'mJy':
            if angle_unit == u.sr:
                ans = 0.96
            else:
                ans = 1.2e+10
            bg_ans = 1e9
        elif flux_unit == 'uJy':
            if angle_unit == u.sr:
                ans = 960
            else:
                ans = 1.2e+13
            bg_ans = 1e12
        elif flux_unit == 'W / (Hz m2)':
            if angle_unit == u.sr:
                ans = 9.6e-30
            else:
                ans = 1.2e-19
            bg_ans = 1e-20
        elif flux_unit == 'eV / (Hz s m2)':
            if angle_unit == u.sr:
                ans = 5.99185e-11
            else:
                ans = 7.48981e-01
            bg_ans = 0.06241509074460764
        elif flux_unit == 'erg / (Angstrom s cm2)':
            if angle_unit == u.sr:
                ans = 1.34580e-15
            else:
                ans = 1.68225e-05
            bg_ans = 1.4018163962192981e-06
        elif flux_unit == 'erg / (Hz s cm2)':
            if angle_unit == u.sr:
                ans = 9.6e-27
            else:
                ans = 1.2e-16
            bg_ans = 1e-17
        elif flux_unit == 'ph / (Angstrom s cm2)':
            if angle_unit == u.sr:
                ans = 3.133e-04
            else:
                ans = 3.91624e+06
            bg_ans = 326346.6709140777
        elif flux_unit == 'ph / (Hz s cm2)':
            if angle_unit == u.sr:
                ans = 2.23486e-15
            else:
                ans = 2.79357e-05
            bg_ans = 2.328027206660126e-06

        uc_plg.flux_unit = flux_unit
        assert cubeviz_helper.app._get_display_unit('spectral_y') == flux_unit

        label_mouseover._viewer_mouse_event(viewer, {
            'event': 'mousemove', 'domain': {'x': 4.6245e-7, 'y': ans}})
        assert_allclose(label_mouseover._dict['value'], ans, rtol=1e-4)

        assert_allclose(ap_plg.background_value, bg_ans, rtol=1e-4)


@pytest.mark.parametrize("flux_unit, expected_choices", [(u.count, ['ct']),
                                                         (u.Jy, SPEC_PHOTON_FLUX_DENSITY_UNITS),
                                                         (u.nJy, SPEC_PHOTON_FLUX_DENSITY_UNITS + ['nJy'])])  # noqa
def test_flux_unit_choices(cubeviz_helper, flux_unit, expected_choices):
    """
    Test that cubes loaded with various flux units have the expected default
    flux unit selection in the unit conversion plugin, and that the list of
    convertable flux units in the dropdown is correct.
    """

    w, wcs_dict = cubeviz_wcs_dict()
    flux = np.zeros((3, 4, 5), dtype=np.float32)
    # load cube in flux_unit, will become cube in flux_unit / pix2
    cube = Spectrum1D(flux=flux * flux_unit, wcs=w, meta=wcs_dict)
    cubeviz_helper.load_data(cube)

    uc_plg = cubeviz_helper.plugins['Unit Conversion']

    assert uc_plg.angle_unit.selected == 'pix2'  # will always be pix2

    assert uc_plg.flux_unit.selected == flux_unit.to_string()
    assert uc_plg.flux_unit.choices == expected_choices


@pytest.mark.parametrize("angle_unit", [u.sr, PIX2])
def test_unit_translation(cubeviz_helper, angle_unit):
    # custom cube so PIXAR_SR is in metadata, and Flux units, and in MJy
    w, wcs_dict = cubeviz_wcs_dict()
    flux = np.zeros((30, 20, 3001), dtype=np.float32)
    flux[5:15, 1:11, :] = 1
    cube = Spectrum1D(flux=flux * u.MJy / angle_unit, wcs=w, meta=wcs_dict)
    cubeviz_helper.load_data(cube, data_label="test")

    center = PixCoord(5, 10)
    cubeviz_helper.plugins['Subset Tools'].import_region(CirclePixelRegion(center, radius=2.5))

    uc_plg = cubeviz_helper.plugins['Unit Conversion']

    # test that the scale factor was set
    assert np.all(cubeviz_helper.app.data_collection['Spectrum (sum)'].meta['_pixel_scale_factor'] != 1)  # noqa

    # When the dropdown is displayed, this ensures the loaded
    # data collection item units will be used for translations.
    assert uc_plg._obj.spectral_y_type_selected == 'Flux'

    # accessing from get_data(use_display_units=True) should return flux-like units
    assert cubeviz_helper.app._get_display_unit('spectral_y') == u.MJy
    assert cubeviz_helper.get_data('Spectrum (sum)', use_display_units=True).unit == u.MJy

    # to have access to display units
    viewer_1d = cubeviz_helper.app.get_viewer(
        cubeviz_helper._default_spectrum_viewer_reference_name)

    # change global y-units from Flux -> Surface Brightness
    uc_plg._obj.spectral_y_type_selected = 'Surface Brightness'

    assert uc_plg._obj.spectral_y_type_selected == 'Surface Brightness'
    y_display_unit = u.Unit(viewer_1d.state.y_display_unit)

    # check if units translated
    assert y_display_unit == u.MJy / angle_unit

    # get_data(use_display_units=True) should return surface brightness-like units
    assert cubeviz_helper.app._get_display_unit('spectral_y') == u.MJy / angle_unit
    assert cubeviz_helper.get_data('Spectrum (sum)', use_display_units=True).unit == u.MJy / angle_unit  # noqa


@pytest.mark.parametrize("angle_unit", [u.sr, PIX2])
def test_sb_unit_conversion(cubeviz_helper, angle_unit):

    angle_str = angle_unit.to_string()

    # custom cube to have Surface Brightness units
    w, wcs_dict = cubeviz_wcs_dict()
    flux = np.zeros((30, 20, 3001), dtype=np.float32)
    flux[5:15, 1:11, :] = 1
    cube = Spectrum1D(flux=flux * (u.MJy / angle_unit), wcs=w, meta=wcs_dict)
    cubeviz_helper.load_data(cube, data_label="test")

    uc_plg = cubeviz_helper.plugins['Unit Conversion']
    uc_plg.open_in_tray()

    # ensure that per solid angle cube defaults to Flux spectrum
    assert uc_plg.spectral_y_type == 'Flux'
    # flux choices is populated with flux units
    assert uc_plg.flux_unit.choices
    # and angle choices should be the only the input angle
    assert len(uc_plg.angle_unit.choices) == 1
    assert angle_str in uc_plg.angle_unit.choices

    # to have access to display units
    viewer_1d = cubeviz_helper.app.get_viewer(
        cubeviz_helper._default_spectrum_viewer_reference_name)

    uc_plg.spectral_y_type.selected = 'Surface Brightness'

    # Surface Brightness conversion
    uc_plg.flux_unit = 'Jy'
    y_display_unit = u.Unit(viewer_1d.state.y_display_unit)
    assert y_display_unit == u.Jy / angle_unit
    label_mouseover = cubeviz_helper.app.session.application._tools["g-coords-info"]
    flux_viewer = cubeviz_helper.app.get_viewer(
        cubeviz_helper._default_flux_viewer_reference_name
    )
    label_mouseover._viewer_mouse_event(
        flux_viewer, {"event": "mousemove", "domain": {"x": 10, "y": 8}}
    )
    assert label_mouseover.as_text() == (
            f"Pixel x=00010.0 y=00008.0 Value +1.00000e+06 Jy / {angle_str}",
            "World 13h39m59.7037s +27d00m03.2400s (ICRS)",
            "204.9987654313 27.0008999946 (deg)")

    # Try a second conversion
    uc_plg.flux_unit = 'W / Hz m2'

    if angle_unit == PIX2:  # unit string order is different for pix2 vs sr
        str_unit = 'W / (Hz m2 pix2)'
    elif angle_unit == u.sr:
        str_unit = 'W / (Hz sr m2)'

    y_display_unit = u.Unit(viewer_1d.state.y_display_unit)
    assert y_display_unit == u.Unit(str_unit)

    y_display_unit = u.Unit(viewer_1d.state.y_display_unit)
    label_mouseover._viewer_mouse_event(
        flux_viewer, {"event": "mousemove", "domain": {"x": 10, "y": 8}}
    )

    assert label_mouseover.as_text() == (
            f"Pixel x=00010.0 y=00008.0 Value +1.00000e-20 {str_unit}",
            "World 13h39m59.7037s +27d00m03.2400s (ICRS)",
            "204.9987654313 27.0008999946 (deg)")

    # really a translation test, test_unit_translation loads a Flux
    # cube, this test load a Surface Brightness Cube, this ensures
    # two-way translation
    uc_plg.flux_unit = 'MJy'
    y_display_unit = u.Unit(viewer_1d.state.y_display_unit)
    label_mouseover._viewer_mouse_event(
        flux_viewer, {"event": "mousemove", "domain": {"x": 10, "y": 8}}
    )
    assert label_mouseover.as_text() == (
            f"Pixel x=00010.0 y=00008.0 Value +1.00000e+00 MJy / {angle_str}",
            "World 13h39m59.7037s +27d00m03.2400s (ICRS)",
            "204.9987654313 27.0008999946 (deg)")

    uc_plg._obj.spectral_y_type_selected = 'Flux'
    uc_plg.flux_unit = 'Jy'
    y_display_unit = u.Unit(viewer_1d.state.y_display_unit)

    assert y_display_unit == u.Jy

    la = cubeviz_helper.plugins['Line Analysis']._obj
    assert la.dataset.get_selected_spectrum(use_display_units=True)


def test_contour_unit_conversion(cubeviz_helper, spectrum1d_cube_fluxunit_jy_per_steradian):
    # custom cube to have Surface Brightness units
    cubeviz_helper.load_data(spectrum1d_cube_fluxunit_jy_per_steradian, data_label="test")

    uc_plg = cubeviz_helper.plugins['Unit Conversion']
    uc_plg.open_in_tray()

    po_plg = cubeviz_helper.plugins['Plot Options']
    # Make sure that the contour values get updated
    po_plg.contour_visible = True

    assert uc_plg.spectral_y_type == 'Flux'
    assert uc_plg.flux_unit == 'Jy'
    assert uc_plg.sb_unit == "Jy / sr"
    assert cubeviz_helper.viewers['flux-viewer']._obj.layers[0].state.attribute_display_unit == "Jy / sr"  # noqa
    assert np.allclose(po_plg.contour_max.value, 199)

    uc_plg.spectral_y_type = 'Surface Brightness'
    uc_plg.flux_unit = 'MJy'

    assert uc_plg.sb_unit == "MJy / sr"
    assert cubeviz_helper.viewers['flux-viewer']._obj.layers[0].state.attribute_display_unit == "MJy / sr"  # noqa
    assert np.allclose(po_plg.contour_max.value, 1.99e-4)


@pytest.mark.parametrize("angle_unit", [u.sr, PIX2])
def test_cubeviz_flux_sb_translation_counts(cubeviz_helper, angle_unit):

    """
    When a cube is loaded in counts, 'count' should be the only
    available option for flux unit. The y axis can be translated
    between flux and sb. Test a flux cube which will be converted
    to ct/pix2, and a sb cube ct/sr.
    """

    angle_str = angle_unit.to_string()

    # custom cube to have Surface Brightness units
    w, wcs_dict = cubeviz_wcs_dict()
    flux = np.zeros((30, 20, 3001), dtype=np.float32)
    flux[5:15, 1:11, :] = 1
    cube = Spectrum1D(flux=flux * (u.ct / angle_unit), wcs=w, meta=wcs_dict)
    cubeviz_helper.load_data(cube, data_label="test")

    uc_plg = cubeviz_helper.plugins['Unit Conversion']
    uc_plg.open_in_tray()

    # ensure that per solid angle cube defaults to Flux spectrum
    assert uc_plg.spectral_y_type == 'Flux'
    # flux choices is populated with only one choice, counts
    assert len(uc_plg.flux_unit.choices) == 1
    assert 'ct' in uc_plg.flux_unit.choices
    # and angle choices should be the only the input angle
    assert len(uc_plg.angle_unit.choices) == 1
    assert angle_str in uc_plg.angle_unit.choices

    # to have access to display units
    viewer_1d = cubeviz_helper.app.get_viewer(
        cubeviz_helper._default_spectrum_viewer_reference_name)

    # do a spectral y axis translation from Flux to Surface Brightness
    uc_plg.spectral_y_type.selected = 'Surface Brightness'

    y_display_unit = u.Unit(viewer_1d.state.y_display_unit)
    assert y_display_unit == u.ct / angle_unit

    # and test mouseover info
    label_mouseover = cubeviz_helper.app.session.application._tools["g-coords-info"]
    flux_viewer = cubeviz_helper.app.get_viewer(
        cubeviz_helper._default_flux_viewer_reference_name
    )
    label_mouseover._viewer_mouse_event(
        flux_viewer, {"event": "mousemove", "domain": {"x": 10, "y": 8}}
    )
    assert label_mouseover.as_text() == (
            f"Pixel x=00010.0 y=00008.0 Value +1.00000e+00 ct / {angle_str}",
            "World 13h39m59.7037s +27d00m03.2400s (ICRS)",
            "204.9987654313 27.0008999946 (deg)")


@pytest.mark.parametrize("start_unit, end_unit, end_spectral_y_type, expected_limits",
                         [('Jy', 'MJy', 'Flux', (5e-07, 6e-07, 1e-4, 1.05e-4)),
                          ('MJy', 'ph / (Angstrom s cm2)', 'Surface Brightness', (5e-07, 6e-07, 25153169.66070254, 31692993.772485193))  # noqa
                          ])
def test_limits_on_unit_change(cubeviz_helper, start_unit, end_unit,
                               end_spectral_y_type, expected_limits):
    """
    Test that the limits are reset when changing units
    """

    w, wcs_dict = cubeviz_wcs_dict()
    flux = np.zeros((30, 20, 3001), dtype=np.float32)
    flux[5:15, 1:11, :] = 1
    cube = Spectrum1D(flux=flux * u.Unit(start_unit), wcs=w, meta=wcs_dict)
    cubeviz_helper.load_data(cube, data_label="test")

    uc_plg = cubeviz_helper.plugins['Unit Conversion']
    sv = cubeviz_helper.viewers['spectrum-viewer']
    sv.set_limits(x_min=5e-7, x_max=6e-7, y_min=100, y_max=105)

    uc_plg.flux_unit = end_unit
    uc_plg.spectral_y_type = end_spectral_y_type

    new_limits = sv._obj.get_limits()
    assert np.allclose(new_limits, expected_limits)



from PyInstaller.utils.hooks import collect_data_files, copy_metadata

datas = collect_data_files('matplotlib_inline')
# since matplotlib 3.9 entry_points.txt is needed
datas += copy_metadata('matplotlib_inline')



from importlib import resources
from pathlib import Path

import numpy as np
from matplotlib.colors import ListedColormap, rgb2hex
from glue.config import stretches
from astropy.table import Table

# paths to CSV files with DQ flag mappings:
dq_flag_map_paths = {
    'jwst': Path('data', 'data_quality', 'jwst.csv'),
    'roman': Path('data', 'data_quality', 'roman.csv'),
    'hst-stis': Path('data', 'data_quality', 'hst-stis.csv'),
    'hst-acs': Path('data', 'data_quality', 'hst-acs.csv'),
    'hst-wfc3-uvis': Path('data', 'data_quality', 'hst-wfc3-uvis.csv'),
    'hst-wfc3-ir': Path('data', 'data_quality', 'hst-wfc3-ir.csv'),
    'hst-cos': Path('data', 'data_quality', 'hst-cos.csv'),
}


class LookupStretch:
    """
    Stretch class specific to DQ arrays.

    Attributes
    ----------
    flags : array-like
        DQ flags.
    """

    def __init__(self, flags=None, hidden_flags=None):
        # Default x, y values(0-1) range chosen for a typical initial spline shape.
        # Can be modified if required.
        if flags is None:
            flags = np.linspace(0, 1, 5)
        if hidden_flags is None:
            hidden_flags = []

        self.flags = np.asarray(flags)
        self.hidden_flags = np.asarray(hidden_flags).astype(int)

    @property
    def flag_range(self):
        return np.max(self.flags) - np.min(self.flags)

    @property
    def scaled_flags(self):
        # renormalize the flags on range (0, 1):
        return (self.flags - np.min(self.flags)) / self.flag_range

    def dq_array_to_flag_index(self, values):
        # Find the index of the closest entry in `scaled_flags`
        # for each of `values` using array broadcasting:
        return np.argmin(
            np.abs(
                np.nan_to_num(values, nan=-10).flatten()[None, :] -
                self.scaled_flags[:, None]
            ), axis=0
        ).reshape(values.shape)

    def __call__(self, values, out=None, clip=False):
        # For our uses, we can ignore `out` and `clip`, but those would need
        # to be implemented before contributing this class upstream.

        # find closest index in `self.flags` for each value in `values`:
        if hasattr(values, 'squeeze'):
            values = values.squeeze()

        # `values` will have already been passed through
        # astropy.visualization.ManualInterval and normalized on (0, 1)
        # before they arrive here. First, remove that interval and get
        # back the integer values:
        values_integer = np.round(values * self.flag_range + np.min(self.flags))

        # normalize by the number of flags, onto interval (0, 1):
        renormed = self.dq_array_to_flag_index(values) / len(self.flags)

        if len(self.hidden_flags):
            # mask that is True for `values` in the hidden flags list:
            value_is_hidden = np.isin(
                np.nan_to_num(values_integer, nan=-10).astype(int),
                self.hidden_flags.astype(int)
            )
        else:
            value_is_hidden = False

        # preserve NaNs in values, and make hidden flags NaNs:
        return np.where(
            np.isnan(values) | value_is_hidden,
            np.nan,
            renormed
        )


if "lookup" not in stretches:
    stretches.add("lookup", LookupStretch, display="DQ")


def load_flag_map(mission_or_instrument=None, path=None):
    """
    Load a flag map from disk.

    Parameters
    ----------
    mission_or_instrument : {"jwst", "roman"}, optional
        Load the flag map for this mission or instrument.
    path : path-like
        Path to a flag map file (CSV).

    Returns
    -------
    flag_mapping : dict
        Flag mapping definitions. Keys are powers of two,
        values are the "name" and "description" for each flag.
    """
    if mission_or_instrument is not None and path is None:
        flag_map_path = dq_flag_map_paths.get(mission_or_instrument)

        if flag_map_path is None:
            raise NotImplementedError("No flag map found for "
                                      f"mission/instrument '{mission_or_instrument}'.")

    elif mission_or_instrument is None and path is not None:
        # load directly from a CSV file:
        flag_map_path = path

    with resources.as_file(resources.files('jdaviz').joinpath(flag_map_path)) as path:
        # by default astropy's Table.read returns "masked" for empty table data, so
        # here we supply a fill value (empty string)
        fill_values = ['']

        flag_table = Table.read(path, format='ascii.csv', fill_values=fill_values)

    flag_mapping = {}
    for flag, name, desc in flag_table[['flag', 'name', 'description']].iterrows():
        flag_mapping[int(flag)] = dict(name=name, description=desc)

    return flag_mapping


def write_flag_map(flag_mapping, csv_path, **kwargs):
    """
    Write a table containing bitmask flags and their descriptions
    to a CSV file.

    Parameters
    ----------
    flag_mapping : dict
        Flag mapping definitions. Keys are powers of two,
        values are the "name" and "description" for each flag.
    csv_path : path-like
        Where to save the CSV output.
    **kwargs : dict
        Other keywords passed to `~astropy.table.Table.write`
    """
    names = ['flag', 'name', 'description']
    dtype = (int, str, str)
    table = Table(names=names, dtype=dtype)

    for key, value in flag_mapping.items():
        row = {'flag': key}
        row.update(value)
        table.add_row(row)

    table.write(csv_path, format='ascii.csv', **kwargs)


def generate_listed_colormap(n_flags=None, rgba_colors=None, seed=3):
    """
    Generate a list of random "light" colors of length ``n_flags``.

    Parameters
    ----------
    n_flags : int
        Number of colors in the listed colormap, should match the
        number of unique DQ flags (before they're decomposed).
    rgba_colors : list of tuples
        List of RGBA tuples for each color in the colormap.
    seed : int
        Seed for the random number generator used to
        draw random colors.

    Returns
    -------
    cmap : `~matplotlib.pyplot.colors.ListedColormap`
        Colormap constructed with ``n_flags`` colors.
    rgba_colors : list of tuples
        Random light colors of length ``n_flags``.
    """
    rng = np.random.default_rng(seed)
    default_alpha = 1

    if rgba_colors is None:
        # Generate random colors that are generally "light", i.e. with
        # RGB values in the upper half of the interval (0, 1):
        rgba_colors = [
            tuple(np.insert(rng.uniform(size=2), rng.integers(0, 3), 1).tolist() + [default_alpha])
            for _ in range(n_flags)
        ]

    cmap = ListedColormap(rgba_colors)

    # setting `bad` alpha=0 will make NaNs transparent:
    cmap.set_bad(color='k', alpha=0)
    return cmap, rgba_colors


def decompose_bit(bit):
    """
    For an integer ``bit``, return a list of the powers of
    two that sum up to ``bit``.

    Parameters
    ----------
    bit : int
        Sum of powers of two.

    Returns
    -------
    powers : list of integers
        Powers of two which sum to ``bit``.
    """
    bit = int(bit)
    powers = []
    i = 1
    while i <= bit:
        if i & bit:
            powers.append(int(np.log2(i)))
        i <<= 1
    return sorted(powers)


def decode_flags(flag_map, unique_flags, rgba_colors):
    """
    For a list of unique bits in ``unique_flags``, return a list of
    dictionaries of the decomposed bits with their names, definitions, and
    colors defined in ``rgba_colors``.

    Parameters
    ----------
    flag_map : dict
        Flag mapping, such as the ones produced by ``load_flag_map``.
    unique_flags : list or array
        Sequence of unique flags which occur in a data quality array.
    rgba_colors : list of tuples
        RGBA color tuples, one per unique flag.
    """
    decoded_flags = []

    for i, (bit, color) in enumerate(zip(unique_flags, rgba_colors)):
        decoded_bits = decompose_bit(bit)
        decoded_flags.append({
            'flag': int(bit),
            'decomposed': {bit: flag_map[bit] if bit in flag_map else bit
                           for bit in decoded_bits},
            'color': rgb2hex(color),
            'show': True,
        })

    return decoded_flags



from traitlets import Bool, Unicode, observe
from urllib.parse import urlparse

from jdaviz.core.registries import loader_resolver_registry
from jdaviz.core.loaders.resolvers import BaseResolver
from jdaviz.core.user_api import LoaderUserApi
from jdaviz.utils import download_uri_to_path


__all__ = ['URLResolver']


@loader_resolver_registry('url')
class URLResolver(BaseResolver):
    template_file = __file__, "url.vue"
    default_input = 'url'

    url = Unicode("").tag(sync=True)
    cache = Bool(True).tag(sync=True)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @property
    def user_api(self):
        return LoaderUserApi(self, expose=['url', 'cache'])

    @property
    def is_valid(self):
        return urlparse(self.url).scheme in ['http', 'https']

    @observe('url', 'cache')
    def _on_url_changed(self, change):
        self._update_format_items()

    def __call__(self, local_path=None, timeout=60):
        return download_uri_to_path(self.url, cache=self.cache,
                                    local_path=local_path, timeout=timeout)



from astropy import units as u
from astropy.coordinates import SkyCoord
from numpy.testing import assert_allclose
from regions import PixCoord, CirclePixelRegion

from jdaviz.configs.imviz.tests.utils import BaseImviz_WCS_GWCS


class TestImvizSpatialSubsetCentroidPixelLinked(BaseImviz_WCS_GWCS):
    def test_centroiding_pixel(self):
        reg = CirclePixelRegion(PixCoord(2, 2), 3)
        plg = self.imviz.plugins['Subset Tools']
        plg.import_region(reg)

        plg.subset = 'Subset 1'

        # Since they are linked by pixels, the bright corner pixel aligns
        # and nothing should change.
        for data_label in ('fits_wcs[DATA]', 'gwcs[DATA]'):
            plg.recenter_dataset = data_label
            plg.set_center((2, 2), update=True)  # Move the Subset back first.
            plg.recenter()

            # Calculate and move to centroid.
            for key in ("X Center (pixels)", "Y Center (pixels)"):
                assert plg._obj._get_value_from_subset_definition(0, key, "value") == -1
                assert plg._obj._get_value_from_subset_definition(0, key, "orig") == -1

            # Radius will not be touched.
            for key in ("value", "orig"):
                assert plg._obj._get_value_from_subset_definition(0, "Radius (pixels)", key) == 3

        assert plg.get_center() == (-1, -1)


class TestImvizSpatialSubsetCentroidWCSLinked(BaseImviz_WCS_GWCS):
    def test_centroiding_wcs(self):
        # FITS WCS and GWCS are rotated from each other.
        # Plain array always by pixel wrt FITS WCS.
        self.imviz.link_data(align_by='wcs')

        reg = CirclePixelRegion(PixCoord(2, 2), 3).to_sky(self.wcs_1)
        plg = self.imviz.plugins['Subset Tools']
        plg.import_region(reg)

        plg.subset = 'Subset 1'
        plg.recenter_dataset = 'fits_wcs[DATA]'
        plg.recenter()

        # Pixel value is now w.r.t. fake WCS layer, not the selected data.
        for key in ("value", "orig"):
            # subset definition is now in sky coordinates. get RA and Dec and convert back to pixel
            # to compare with expected recentered position.
            ra = plg._obj._get_value_from_subset_definition(0, "RA Center (degrees)", key) * u.deg
            dec = plg._obj._get_value_from_subset_definition(0, "Dec Center (degrees)", key) * u.deg
            x, y = SkyCoord(ra, dec).to_pixel(self.wcs_1)
            assert_allclose((x, y), -1)

        # GWCS does not extrapolate and this Subset is out of bounds,
        # so will get NaNs and enter the exception handling logic.
        plg.recenter_dataset = 'gwcs[DATA]'
        plg.set_center((2.6836, 1.6332), update=True)  # Move the Subset back first.
        plg.recenter()
        subsets = self.imviz.app.get_subsets(include_sky_region=True)
        subsets_sky = subsets['Subset 1'][0]['sky_region']
        subsets_pix = subsets['Subset 1'][0]['region']
        assert_allclose((subsets_pix.center.x, subsets_pix.center.y), (2.6836, 1.6332))
        for key in ("value", "orig"):
            ra = plg._obj._get_value_from_subset_definition(0, "RA Center (degrees)", key)
            dec = plg._obj._get_value_from_subset_definition(0, "Dec Center (degrees)", key)

            # make sure what is in subset_definitions matches what is returned by get_subsets
            assert_allclose((ra, dec), (subsets_sky.center.ra.deg, subsets_sky.center.dec.deg))

        # The functionality for set_center has changed so that the subset state itself
        # is updated but that change is not propagated to subset_definitions or the UI until
        # vue_update_subset is called.
        plg._obj.set_center((2, 2), update=False)
        for key in ("value", "orig"):
            ra = plg._obj._get_value_from_subset_definition(0, "RA Center (degrees)", key)
            dec = plg._obj._get_value_from_subset_definition(0, "Dec Center (degrees)", key)
            # here 'ra' and 'dec' should remain unchanged from when they were defined, since
            # vue_update_subset hasn't run
            assert_allclose((ra, dec), (subsets_sky.center.ra.deg, subsets_sky.center.dec.deg))



"""
Helper classes are meant to provide a convenient user API for specific
configurations. They allow a separation of "viztool-specific" API and the glue
application objects.

See also https://github.com/spacetelescope/jdaviz/issues/104 for more details
on the motivation behind this concept.
"""
import warnings
from contextlib import contextmanager
from inspect import isclass

import numpy as np
from glue.core import HubListener
from glue.core.edit_subset_mode import NewMode
from glue.core.message import SubsetCreateMessage, SubsetDeleteMessage
from glue.core.subset import Subset, MaskSubsetState
from glue.config import data_translator
from ipywidgets.widgets import widget_serialization

from astropy.nddata import NDDataArray, CCDData, StdDevUncertainty
import astropy.units as u
from astropy.utils.decorators import deprecated
from regions.core.core import Region
from specutils import Spectrum1D, SpectralRegion

from jdaviz.app import Application
from jdaviz.core.events import SnackbarMessage, ExitBatchLoadMessage, SliceSelectSliceMessage
from jdaviz.core.loaders.resolvers import find_matching_resolver
from jdaviz.core.template_mixin import show_widget
from jdaviz.utils import data_has_valid_wcs
from jdaviz.core.unit_conversion_utils import (all_flux_unit_conversion_equivs,
                                               check_if_unit_is_per_solid_angle,
                                               flux_conversion_general,
                                               spectral_axis_conversion)


__all__ = ['ConfigHelper', 'ImageConfigHelper', 'CubeConfigHelper']


class ConfigHelper(HubListener):
    """The Base Helper Class.
    Provides shared abstracted helper methods to the user.

    Subclasses should set ``_default_configuration`` if they are meant to be
    used with a specific configuration.

    Parameters
    ----------
    app : `~jdaviz.app.Application` or `None`
        The application object, or if `None`, creates a new one based on the
        default configuration for this helper.
    verbosity : {'debug', 'info', 'warning', 'error'}
        Verbosity of the popup messages in the application.
    history_verbosity : {'debug', 'info', 'warning', 'error'}
        Verbosity of the history logger in the application.
    """
    _default_configuration = 'default'

    def __init__(self, app=None, verbosity='warning', history_verbosity='info'):
        if app is None:
            self.app = Application(configuration=self._default_configuration)
        else:
            self.app = app
        self.app.verbosity = verbosity
        self.app.history_verbosity = history_verbosity

        # give a reference from the app back to this config helper.  These can be accessed from a
        # viewer via viewer.jdaviz_app and viewer.jdaviz_helper
        # if the helper has already been set, this is probably a nested viz tool. Don't overwrite
        if self.app._jdaviz_helper is None:
            self.app._jdaviz_helper = self

        self.app.hub.subscribe(self, SubsetCreateMessage,
                               handler=lambda msg: self._propagate_callback_to_viewers('_on_subset_create', msg)) # noqa
        self.app.hub.subscribe(self, SubsetDeleteMessage,
                               handler=lambda msg: self._propagate_callback_to_viewers('_on_subset_delete', msg)) # noqa

        self._in_batch_load = 0
        self._delayed_show_in_viewer_labels = {}  # label: viewer_reference pairs

    def _propagate_callback_to_viewers(self, method, msg):
        # viewers don't have access to the app/hub to subscribe to messages, so we'll
        # catch all messages here and pass them on to each of the viewers that
        # have the applicable method implemented.
        for viewer in self.app._viewer_store.values():
            if hasattr(viewer, method):
                getattr(viewer, method)(msg)

    @contextmanager
    def batch_load(self):
        """
        Context manager to delay linking and loading data into viewers
        """
        # we'll use a counter instead of a boolean to allow the user to nest multiple
        # context managers.  Once they're all exited, then the linking/showing will
        # take place.
        self._in_batch_load += 1
        with self.app.data_collection.delay_link_manager_update():
            # user entrypoint (anything within the with-statement will get called here)
            yield

        self._in_batch_load -= 1
        if not self._in_batch_load:
            self.app.hub.broadcast(ExitBatchLoadMessage(sender=self.app))

            # add any data to viewers that were requested but deferred
            for data_label, viewer_ref in self._delayed_show_in_viewer_labels.items():
                self.app.set_data_visibility(viewer_ref, data_label,
                                             visible=True, replace=False)
            self._delayed_show_in_viewer_labels = {}

    def load_data(self, data, data_label=None, parser_reference=None, **kwargs):
        if data_label:
            kwargs['data_label'] = data_label
        self.app.load_data(data, parser_reference=parser_reference, **kwargs)

    @property
    def loaders(self):
        """
        Access API objects for data loaders in the import dialog.

        Returns
        -------
        loaders : dict
            dict of loader objects
        """
        if not self.app.state.dev_loaders:
            raise NotImplementedError("loaders is under active development and requires a dev-flag to test")  # noqa
        loaders = {item['label']: widget_serialization['from_json'](item['widget'], None).user_api
                   for item in self.app.state.loader_items}
        return loaders

    def _load(self, inp=None, loader=None, format=None, target=None, **kwargs):
        """
        Load data into the app.  A single valid loader/importer must be able to be
        matched based on the input, otherwise an error will be raised suggesting
        what further information to provide.  For an interactive approach,
        see ``loaders``.

        Parameters
        ----------
        inp : string or object or None
            Input filename, url, data object, etc.
        loader : string, optional
            Only consider a specific loader/resolver
        format : string, optional
            Only consider a specific format
        target : string, optional
            Only consider a specific target
        kwargs :
            Additional kwargs are passed on to both the loader and importer, as applicable.
            Any kwargs that do not match valid inputs are silently ignored.
        """
        resolver = find_matching_resolver(self.app, inp,
                                          resolver=loader,
                                          format=format,
                                          target=target,
                                          **kwargs)

        importer = resolver.importer
        for k, v in kwargs.items():
            if hasattr(importer, k):
                setattr(importer, k, v)
        return importer()

    @property
    def data_labels(self):
        """
        List of data labels loaded and available in jdaviz

        Returns
        -------
        data_labels : list
            list of strings
        """
        return [data.label for data in self.app.data_collection]

    @property
    def plugins(self):
        """
        Access API objects for plugins in the plugin tray.

        Returns
        -------
        plugins : dict
            dict of plugin objects
        """
        plugins = {item['label']: widget_serialization['from_json'](item['widget'], None).user_api
                   for item in self.app.state.tray_items if item['is_relevant']}

        # handle renamed plugins during deprecation
        if 'Image Profiles (XY)' in plugins:
            # renamed in 4.0
            plugins['Imviz Line Profiles (XY)'] = plugins['Image Profiles (XY)']._obj.user_api
            plugins['Imviz Line Profiles (XY)']._deprecation_msg = 'in the future, the formerly named \"Imviz Line Profiles (XY)\" plugin will only be available by its new name: \"Image Profiles (XY)\".'  # noqa

        return plugins

    @property
    def plugin_tables(self):
        return self.app._plugin_tables

    @property
    def plugin_plots(self):
        return self.app._plugin_plots

    @property
    def viewers(self):
        """
        Access API objects for any viewer.

        Returns
        -------
        viewers : dict
            dict of viewer objects
        """
        return {viewer._ref_or_id: viewer.user_api
                for viewer in self.app._viewer_store.values()}

    @property
    @deprecated(since="4.2", alternative="plugins['Model Fitting'].fitted_models")
    def fitted_models(self):
        """
        Returns the fitted models.

        Returns
        -------
        parameters : dict
            dict of `astropy.modeling.Model` objects, or None.
        """
        plg = self.plugins.get('Model Fitting', None)
        if plg is None:
            raise ValueError("Model Fitting plugins is not loaded")
        return plg.fitted_models

    @deprecated(since="4.2", alternative="plugins['Model Fitting'].get_models")
    def get_models(self, models=None, model_label=None, x=None, y=None):
        """
        Loop through all models and output models of the label model_label.
        If x or y is set, return model_labels of those (x, y) coordinates.
        If x and y are None, print all models regardless of coordinates.

        Parameters
        ----------
        models : dict
            A dict of models, with the key being the label name and the value
            being an `astropy.modeling.CompoundModel` object. Defaults to
            `fitted_models` if no parameter is provided.
        model_label : str
            The name of the model that will be found and returned. If it
            equals default, every model present will be returned.
        x : int
            The x coordinate of the model spaxels that will be returned.
        y : int
            The y coordinate of the model spaxels that will be returned.

        Returns
        -------
        selected_models : dict
            Dictionary of the selected models.
        """
        plg = self.plugins.get('Model Fitting', None)
        if plg is None:
            raise ValueError("Model Fitting plugins is not loaded")
        return plg._obj.get_models(models=models,
                                   model_label=model_label,
                                   x=x, y=y)

    @deprecated(since="4.2")
    def get_model_parameters(self, models=None, model_label=None, x=None, y=None):
        """
        Convert each parameter of model inside models into a coordinate that
        maps the model name and parameter name to a `astropy.units.Quantity`
        object.

        Parameters
        ----------
        models : dict
            A dictionary where the key is a model name and the value is an
            `astropy.modeling.CompoundModel` object.
        model_label : str
            Get model parameters for a particular model by inputting its label.
        x : int
            The x coordinate of the model spaxels that will be returned from
            get_models.
        y : int
            The y coordinate of the model spaxels that will be returned from
            get_models.

        Returns
        -------
        :dict: a dictionary of the form
            {model name: {parameter name: [[`astropy.units.Quantity`]]}}
            for 3d models or
            {model name: {parameter name: `astropy.units.Quantity`}} where the
            Quantity object represents the parameter value and unit of one of
            spaxel models or the 1d models, respectively.
        """
        plg = self.plugins.get('Model Fitting', None)
        if plg is None:
            raise ValueError("Model Fitting plugins is not loaded")
        return plg._obj.get_model_parameters(models=models,
                                             model_label=model_label,
                                             x=x, y=y)

    def show(self, loc="inline", title=None, height=None):  # pragma: no cover
        """Display the Jdaviz application.

        Parameters
        ----------
        loc : str
            The display location determines where to present the viz app.
            Supported locations:

            "inline": Display the Jdaviz application inline in a notebook.
            Note this is functionally equivalent to displaying the cell
            ``viz.app`` in the notebook.

            "sidecar": Display the Jdaviz application in a separate JupyterLab window from the
            notebook, the location of which is decided by the 'anchor.' right is the default

                Other anchors:

                * ``sidecar:right`` (The default, opens a tab to the right of display)
                * ``sidecar:tab-before`` (Full-width tab before the current notebook)
                * ``sidecar:tab-after`` (Full-width tab after the current notebook)
                * ``sidecar:split-right`` (Split-tab in the same window right of the notebook)
                * ``sidecar:split-left`` (Split-tab in the same window left of the notebook)
                * ``sidecar:split-top`` (Split-tab in the same window above the notebook)
                * ``sidecar:split-bottom`` (Split-tab in the same window below the notebook)

                See `jupyterlab-sidecar <https://github.com/jupyter-widgets/jupyterlab-sidecar>`_
                for the most up-to-date options.

            "popout": Display the Jdaviz application in a detached display. By default, a new
            window will open. Browser popup permissions required.

                Other anchors:

                * ``popout:window`` (The default, opens Jdaviz in a new, detached popout)
                * ``popout:tab`` (Opens Jdaviz in a new, detached tab in your browser)

        title : str, optional
            The title of the sidecar tab.  Defaults to the name of the
            application; e.g., "specviz".

            NOTE: Only applicable to a "sidecar" display.

        height: int, optional
            The height of the top-level application widget, in pixels. Applies to all
            instances of the same application in the notebook.

        Notes
        -----
        If "sidecar" is requested in the "classic" Jupyter notebook, the app will appear inline,
        as only JupyterLab has a mechanism to have multiple tabs.
        """
        title = self.app.config if title is None else title
        if height is not None:
            if isinstance(height, int):
                height = f"{height}px"
            self.app.layout.height = height
            self.app.state.settings['context']['notebook']['max_height'] = height

        show_widget(self.app, loc=loc, title=title)

    def show_in_sidecar(self, anchor=None, title=None):  # pragma: no cover
        """
        Preserved for backwards compatibility
        Shows Jdaviz in a sidecar with the default anchor: right
        """
        warnings.warn('show_in_sidecar has been replaced with show(loc="sidecar")',
                      DeprecationWarning)
        location = 'sidecar' if anchor is None else f"sidecar:{anchor}"
        return self.show(loc=location, title=title)

    def show_in_new_tab(self, title=None):  # pragma: no cover
        """
        Preserved for backwards compatibility
        Shows Jdaviz in a sidecar in a new tab to the right
        """
        warnings.warn('show_in_new_tab has been replaced with show(loc="sidecar:tab-after")',
                      DeprecationWarning)
        return self.show(loc="sidecar:tab-after", title=title)

    def toggle_api_hints(self, enabled=None):
        """
        Toggle the visibility of API hints in the application.

        Parameters
        ----------
        enabled : bool, optional
            If `True`, show API hints. If `False`, hide API hints.
            If `None`, toggle the current state.
        """
        if enabled is None:
            enabled = not self.app.state.show_api_hints
        self.app.state.show_api_hints = enabled

    def _handle_display_units(self, data, use_display_units=True):
        if use_display_units:
            if isinstance(data, Spectrum1D):
                spectral_unit = self.app._get_display_unit('spectral')
                if not spectral_unit:
                    return data
                y_unit = self.app._get_display_unit('spectral_y')

                # if there is no pixel scale factor, and the requested conversion
                # is between flux/sb, then skip. this case is encountered when
                # starting the app? ideally this should raise an error, and this
                # should allow pix2/spaxel but it doesn't - keeping this
                # condition as-is until further investigation
                orig_sa = check_if_unit_is_per_solid_angle(u.Unit(data.flux.unit))
                targ_sa = check_if_unit_is_per_solid_angle(u.Unit(y_unit))
                skip_flux_conv = ('_pixel_scale_factor' not in data.meta) & (orig_sa != targ_sa)

                # equivalencies for flux/sb unit conversions
                pixar_sr = data.meta.get('_pixel_scale_factor', None)
                eqv = all_flux_unit_conversion_equivs(pixar_sr=pixar_sr,
                                                      cube_wave=data.spectral_axis)

                # TODO: any other attributes (meta, wcs, etc)?
                # TODO: implement uncertainty.to upstream
                uncertainty = data.uncertainty
                if uncertainty is not None:
                    # convert the uncertainties to StdDevUncertainties, since
                    # that is assumed in a few places in jdaviz:
                    if uncertainty.unit is None:
                        uncertainty.unit = data.flux.unit
                    if hasattr(uncertainty, 'represent_as'):
                        new_uncert = uncertainty.represent_as(StdDevUncertainty)
                    else:
                        # if not specified as NDUncertainty, assume stddev:
                        new_uncert = uncertainty

                    # convert uncertainty units to display units
                    if skip_flux_conv:
                        new_uncert = StdDevUncertainty(new_uncert, unit=data.flux.unit)
                    else:
                        new_uncert_conv = flux_conversion_general(new_uncert.quantity.value,
                                                                  new_uncert.unit,
                                                                  y_unit,
                                                                  eqv,
                                                                  with_unit=False)
                        new_uncert = StdDevUncertainty(new_uncert_conv,
                                                       unit=y_unit)
                else:
                    new_uncert = None

                # convert flux/sb units to display units
                if skip_flux_conv:
                    new_y = data.flux.value * u.Unit(data.flux.unit)
                else:
                    # multiply by unit rather than using with_unit because of
                    # edge case for dimensionless we want here
                    new_y = flux_conversion_general(data.flux.value,
                                                    data.flux.unit,
                                                    y_unit,
                                                    eqv, with_unit=False) * u.Unit(y_unit)

                # convert spectral axis to display units
                new_spec = (spectral_axis_conversion(data.spectral_axis.value,
                                                     data.spectral_axis.unit,
                                                     spectral_unit)
                            * u.Unit(spectral_unit))

                data = Spectrum1D(spectral_axis=new_spec,
                                  flux=new_y,
                                  uncertainty=new_uncert,
                                  mask=data.mask)
            else:  # pragma: nocover
                raise NotImplementedError(f"converting {data.__class__.__name__} to display units is not supported")  # noqa
        return data

    def _get_data(self, data_label=None, spatial_subset=None, spectral_subset=None,
                  temporal_subset=None, mask_subset=None, cls=None, use_display_units=False):
        list_of_valid_subset_names = [x.label for x in self.app.data_collection.subset_groups]
        for subset in (spatial_subset, spectral_subset, mask_subset):
            if subset and subset not in list_of_valid_subset_names:
                raise ValueError(f"Subset {subset} not in list of valid"
                                 f" subset names {list_of_valid_subset_names}")

        if data_label and data_label not in self.app.data_collection.labels:
            raise ValueError(f'{data_label} not in {self.app.data_collection.labels}.')
        elif not data_label and len(self.app.data_collection) > 1:
            raise ValueError('data_label must be set if more than'
                             ' one data exists in data_collection.')
        elif not data_label and len(self.app.data_collection) == 1:
            data_label = self.app.data_collection[0].label

        if cls is not None and not isclass(cls):
            raise TypeError(
                "cls in get_data must be a class or None.")

        if spectral_subset:
            if mask_subset is not None:
                raise ValueError("cannot use both mask_subset and spectral_subset")
            # spectral_subset is applied as a mask, the only difference is that it has
            # its own set of validity checks (whereas mask_subset can be used by downstream
            # apps which would then need to do their own type checks, if necessary)
            mask_subset = spectral_subset

        if temporal_subset:
            if mask_subset is not None:
                raise ValueError("cannot use both mask_subset and spectral_subset")
            mask_subset = temporal_subset

        # End validity checks and start data retrieval
        data = self.app.data_collection[data_label]

        if not cls:
            if 'Trace' in data.meta:
                cls = None
            elif data.ndim == 2 and self.app.config == "specviz2d":
                cls = Spectrum1D
            elif data.ndim == 2:
                cls = CCDData
            elif data.ndim in [1, 3]:
                if self.app.config == 'rampviz':
                    cls = NDDataArray
                else:
                    # for cubeviz, specviz, mosviz, this must be a spectrum:
                    cls = Spectrum1D

        object_kwargs = {}
        if cls == Spectrum1D:
            object_kwargs['statistic'] = None

        if not spatial_subset and not mask_subset:
            if 'Trace' in data.meta:
                if cls is not None:  # pragma: no cover
                    raise ValueError("cls not supported for Trace object")
                data = data.get_object()
            else:
                data = data.get_object(cls=cls, **object_kwargs)

            return self._handle_display_units(data, use_display_units)

        if not cls and spatial_subset:
            raise AttributeError(f"A valid cls must be provided to"
                                 f" apply subset {spatial_subset} to data. "
                                 f"Instead, {cls} was given.")
        elif not cls and mask_subset:
            raise AttributeError(f"A valid cls must be provided to"
                                 f" apply subset {mask_subset} to data. "
                                 f"Instead, {cls} was given.")

        # Now we work on applying subsets to the data
        all_subsets = self.app.get_subsets(object_only=True)

        # Handle spatial subset
        if spatial_subset and not isinstance(all_subsets[spatial_subset][0],
                                             Region):
            raise ValueError(f"{spatial_subset} is not a spatial subset.")
        elif spatial_subset:
            real_spatial = [sub for subsets in self.app.data_collection.subset_groups
                            for sub in subsets.subsets
                            if sub.data.label == data_label and subsets.label == spatial_subset][0]
            handler, _ = data_translator.get_handler_for(cls)
            try:
                data = handler.to_object(real_spatial, **object_kwargs)
            except Exception as e:
                warnings.warn(f"Not able to get {data_label} returned with"
                              f" subset {spatial_subset} applied of type {cls}."
                              f" Exception: {e}")

        # Handle spectral subset, including case where spatial subset is also set
        if spectral_subset and not isinstance(all_subsets[spectral_subset],
                                              SpectralRegion):
            raise ValueError(f"{spectral_subset} is not a spectral subset.")

        if mask_subset:
            real_spectral = [sub for subsets in self.app.data_collection.subset_groups
                             for sub in subsets.subsets
                             if sub.data.label == data_label and subsets.label == mask_subset][0] # noqa

            handler, _ = data_translator.get_handler_for(cls)
            try:
                spec_subset = handler.to_object(real_spectral, **object_kwargs)
            except Exception as e:
                warnings.warn(f"Not able to get {data_label} returned with"
                              f" subset {mask_subset} applied of type {cls}."
                              f" Exception: {e}")
            if spatial_subset:
                # Return collapsed Spectrum1D object with spectral subset mask applied
                data.mask = spec_subset.mask
            else:
                data = spec_subset

        return self._handle_display_units(data, use_display_units)

    def get_data(self, data_label=None, cls=None, use_display_units=False, **kwargs):
        """
        Returns data with name equal to data_label of type cls.

        Parameters
        ----------
        data_label : str, optional
            Provide a label to retrieve a specific data set from data_collection.
        cls : `~specutils.Spectrum1D`, `~astropy.nddata.CCDData`, optional
            The type that data will be returned as.
        use_display_units : bool, optional
            Whether to convert to the display units defined in the <unit-conversion> plugin.

        Returns
        -------
        data : cls
            Data is returned as type ``cls``.

        """
        return self._get_data(data_label=data_label,
                              cls=cls, use_display_units=use_display_units)


class ImageConfigHelper(ConfigHelper):
    """`ConfigHelper` that uses an image viewer as its primary viewer.
    For example, Imviz and Cubeviz.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # NOTE: The first viewer must always be there and is an image viewer.
        self._default_viewer = self.app.get_viewer_by_id(f'{self.app.config}-0')

    @property
    def default_viewer(self):
        """Default viewer instance. This is typically the first viewer
        (e.g., "imviz-0" or "cubeviz-0")."""
        return self._default_viewer.user_api

    @deprecated(since="4.2", alternative="subset_tools.import_region")
    def load_regions_from_file(self, region_file, region_format='ds9', max_num_regions=20,
                               **kwargs):
        """Load regions defined in the given file.
        See :ref:`regions:regions_io` for supported file formats.
        See :meth:`load_regions` for how regions are loaded.

        Parameters
        ----------
        region_file : str
            Path to region file.

        region_format : {'crtf', 'ds9', 'fits'}
            See :meth:`regions.Regions.get_formats`.

        max_num_regions : int or `None`
            Maximum number of regions to read from the file, starting
            from top of the file. Default is first 20 regions that
            can be successfully loaded. If `None`, it will load everything.

        kwargs : dict
            See :meth:`load_regions`.

        Returns
        -------
        bad_regions : list of (obj, str) or `None`
            See :meth:`load_regions`.

        """
        from regions import Regions
        raw_regs = Regions.read(region_file, format=region_format)
        return self.load_regions(raw_regs, max_num_regions=max_num_regions, **kwargs)

    @deprecated(since="4.2", alternative="subset_tools.import_region")
    def load_regions(self, regions, max_num_regions=None, refdata_label=None,
                     return_bad_regions=False, **kwargs):
        """Load given region(s) into the viewer.
        WCS-to-pixel translation and mask creation, if needed, is relative
        to the image defined by ``refdata_label``. Meanwhile, the rest of
        the Subset operations are based on reference image as defined by Glue.

        .. note:: Loading too many regions will affect performance.

        A valid region can be loaded into one of the following categories:

        * An interactive Subset, as if it was drawn by hand. This is
          always done for supported shapes. Its label will be
          ``'Subset N'``, where ``N`` is an integer.
        * A masked Subset that will display on the image but cannot be
          modified once loaded. This is done if the shape cannot be
          made interactive. Its label will be ``'MaskedSubset N'``,
          where ``N`` is an integer.

        Parameters
        ----------
        regions : list of obj
            A list of region objects. A region object can be one of the following:

            * Astropy ``regions`` object
            * ``photutils`` apertures (limited support until ``photutils``
              fully supports ``regions``)

        max_num_regions : int or `None`
            Maximum number of regions to load, starting from top of the list.
            Default is to load everything.

        refdata_label : str or `None`
            Label of data to use for sky-to-pixel conversion for a region, or
            mask creation. Data must already be loaded into Jdaviz.
            If `None`, defaults to the reference data in the default viewer.
            Choice of this data is particularly important when sky
            region is involved.

        return_bad_regions : bool
            If `True`, return the regions that failed to load (see ``bad_regions``);
            This is useful for debugging. If `False`, do not return anything (`None`).

        kwargs : dict
            Extra keywords to be passed into the region's ``to_mask`` method.
            **This is ignored if the region can be made interactive.**

        Returns
        -------
        bad_regions : list of (obj, str) or `None`
            If requested (see ``return_bad_regions`` option), return a
            list of ``(region, reason)`` tuples for region objects that failed to load.
            If all the regions loaded successfully, this list will be empty.
            If not requested, return `None`.

        """
        if len(self.app.data_collection) == 0:
            raise ValueError('Cannot load regions without data.')

        from photutils.aperture import (CircularAperture, SkyCircularAperture,
                                        EllipticalAperture, SkyEllipticalAperture,
                                        RectangularAperture, SkyRectangularAperture,
                                        CircularAnnulus, SkyCircularAnnulus)
        from regions import (Regions, CirclePixelRegion, CircleSkyRegion,
                             EllipsePixelRegion, EllipseSkyRegion,
                             RectanglePixelRegion, RectangleSkyRegion,
                             CircleAnnulusPixelRegion, CircleAnnulusSkyRegion)
        from jdaviz.core.region_translators import regions2roi, aperture2regions

        # If user passes in one region obj instead of list, try to be smart.
        if not isinstance(regions, (list, tuple, Regions)):
            regions = [regions]

        n_loaded = 0
        bad_regions = []

        # To keep track of masked subsets.
        msg_prefix = 'MaskedSubset'
        msg_count = _next_subset_num(msg_prefix, self.app.data_collection.subset_groups)

        # Subset is global but reference data is viewer-dependent.
        if refdata_label is None:
            data = self.default_viewer._obj.state.reference_data
        else:
            data = self.app.data_collection[refdata_label]

        # TODO: Make this work for data cube.
        # https://github.com/glue-viz/glue-astronomy/issues/75
        has_wcs = data_has_valid_wcs(data, ndim=2)

        for region in regions:
            if (isinstance(region, (SkyCircularAperture, SkyEllipticalAperture,
                                    SkyRectangularAperture, SkyCircularAnnulus,
                                    CircleSkyRegion, EllipseSkyRegion,
                                    RectangleSkyRegion, CircleAnnulusSkyRegion))
                    and not has_wcs):
                bad_regions.append((region, 'Sky region provided but data has no valid WCS'))
                continue

            if (isinstance(region, (CircularAperture, EllipticalAperture,
                                    RectangularAperture, CircularAnnulus,
                                    CirclePixelRegion, EllipsePixelRegion,
                                    RectanglePixelRegion, CircleAnnulusPixelRegion))
                    and self.app._align_by == "wcs"):
                bad_regions.append((region, 'Pixel region provided by data is linked by WCS'))
                continue

            # photutils: Convert to regions shape first
            if isinstance(region, (CircularAperture, SkyCircularAperture,
                                   EllipticalAperture, SkyEllipticalAperture,
                                   RectangularAperture, SkyRectangularAperture,
                                   CircularAnnulus, SkyCircularAnnulus)):
                region = aperture2regions(region)

            # regions: Convert to ROI.
            # NOTE: Out-of-bounds ROI will succeed; this is native glue behavior.
            if isinstance(region, (CirclePixelRegion, CircleSkyRegion,
                                   EllipsePixelRegion, EllipseSkyRegion,
                                   RectanglePixelRegion, RectangleSkyRegion,
                                   CircleAnnulusPixelRegion, CircleAnnulusSkyRegion)):
                state = regions2roi(region, wcs=data.coords)

                # TODO: Do we want user to specify viewer? Does it matter?
                self.app.session.edit_subset_mode._mode = NewMode
                self.default_viewer._obj.apply_roi(state)
                self.app.session.edit_subset_mode.edit_subset = None  # No overwrite next iteration # noqa

            # Last resort: Masked Subset that is static (if data is not a cube)
            elif data.ndim == 2:
                im = None
                if hasattr(region, 'to_pixel'):  # Sky region: Convert to pixel region
                    if not has_wcs:
                        bad_regions.append((region, 'Sky region provided but data has no valid WCS'))  # noqa
                        continue
                    region = region.to_pixel(data.coords)

                if hasattr(region, 'to_mask'):
                    try:
                        mask = region.to_mask(**kwargs)
                        im = mask.to_image(data.shape)  # Can be None
                    except Exception as e:  # pragma: no cover
                        bad_regions.append((region, f'Failed to load: {repr(e)}'))
                        continue

                # Boolean mask as input is supported but not advertised.
                elif (isinstance(region, np.ndarray) and region.shape == data.shape
                        and region.dtype == np.bool_):
                    im = region

                if im is None:
                    bad_regions.append((region, 'Mask creation failed'))
                    continue

                # NOTE: Region creation info is thus lost.
                try:
                    subset_label = f'{msg_prefix} {msg_count}'
                    state = MaskSubsetState(im, data.pixel_component_ids)
                    self.app.data_collection.new_subset_group(subset_label, state)
                    msg_count += 1
                except Exception as e:  # pragma: no cover
                    bad_regions.append((region, f'Failed to load: {repr(e)}'))
                    continue

            else:
                bad_regions.append((region, 'Mask creation failed'))
                continue

            n_loaded += 1
            if max_num_regions is not None and n_loaded >= max_num_regions:
                break

        n_reg_in = len(regions)
        n_reg_bad = len(bad_regions)
        if n_loaded == 0:
            snack_color = "error"
        elif n_reg_bad > 0:
            snack_color = "warning"
        else:
            snack_color = "success"
        self.app.hub.broadcast(SnackbarMessage(
            f"Loaded {n_loaded}/{n_reg_in} regions, max_num_regions={max_num_regions}, "
            f"bad={n_reg_bad}", color=snack_color, timeout=8000, sender=self.app))

        if return_bad_regions:
            return bad_regions

    @deprecated(since="4.1", alternative="subset_tools.get_regions")
    def get_interactive_regions(self):
        """
        Return spatial regions that can be interacted with in the viewer.
        This does not return masked regions added via :meth:`load_regions`.

        Unsupported region shapes will be skipped. When that happens,
        a yellow snackbar message will appear on display.

        Returns
        -------
        regions : dict
            Dictionary mapping interactive region names to respective Astropy
            ``regions`` objects.

        """
        from glue_astronomy.translators.regions import roi_subset_state_to_region

        regions = {}
        failed_regs = set()
        to_sky = self.app._align_by == 'wcs'

        # Subset is global, so we just use default viewer.
        for lyr in self.default_viewer._obj.layers:
            if (not hasattr(lyr, 'layer') or not isinstance(lyr.layer, Subset)
                    or lyr.layer.ndim not in (2, 3)):
                continue

            subset_data = lyr.layer
            subset_label = subset_data.label

            try:
                if self.app.config == "imviz" and to_sky:
                    region = roi_subset_state_to_region(subset_data.subset_state, to_sky=to_sky)
                else:
                    region = subset_data.data.get_selection_definition(
                        subset_id=subset_label, format='astropy-regions')
            except (NotImplementedError, ValueError):
                failed_regs.add(subset_label)
            else:
                regions[subset_label] = region

        if len(failed_regs) > 0:
            self.app.hub.broadcast(SnackbarMessage(
                f"Regions skipped: {', '.join(sorted(failed_regs))}",
                color="warning", timeout=8000, sender=self.app))

        return regions

    # TODO: Make this public API?
    def _delete_region(self, subset_label):
        """Delete region given the Subset label."""
        all_subset_labels = [s.label for s in self.app.data_collection.subset_groups]
        if subset_label not in all_subset_labels:
            return
        i = all_subset_labels.index(subset_label)
        subset_grp = self.app.data_collection.subset_groups[i]
        self.app.data_collection.remove_subset_group(subset_grp)


def _next_subset_num(label_prefix, subset_groups):
    """Assumes ``prefix i`` format.
    Does not go back and fill in lower but available numbers. This is consistent with Glue.
    """
    max_i = 0

    for sg in subset_groups:
        if sg.label.startswith(label_prefix):
            sub_label = sg.label.split(' ')
            if len(sub_label) > 1:
                i_str = sub_label[-1]
                try:
                    i = int(i_str)
                except Exception:  # nosec
                    continue
                else:
                    if i > max_i:
                        max_i = i

    return max_i + 1


class CubeConfigHelper(ImageConfigHelper):
    """Base config helper class for cubes"""

    _loaded_flux_cube = None
    _loaded_uncert_cube = None

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @deprecated(since="4.1", alternative="plugins['Slice'].slice")
    def select_slice(self, value):
        """
        Select the slice closest to the provided value.

        Parameters
        ----------
        value : float or int, optional
            Slice value to select in units of the x-axis of the profile viewer.
            The nearest slice will be selected if "snap to slice" is enabled in
            the slice plugin.
        """
        msg = SliceSelectSliceMessage(value=value, sender=self)
        self.app.hub.broadcast(msg)



import re

from glue.config import DictRegistry
from ipyvuetify import VuetifyTemplate
from ipywidgets import Widget


__all__ = ['convert', 'UniqueDictRegistry', 'ViewerRegistry', 'TrayRegistry',
           'ToolRegistry', 'MenuRegistry', 'DataParserRegistry',
           'viewer_registry', 'tray_registry', 'tool_registry', 'menu_registry',
           'data_parser_registry',
           'loader_resolver_registry', 'loader_parser_registry', 'loader_importer_registry']


def _to_snake(s):
    """Convert dashes in viewer category names to underscores for
    use in class attributes, constructor kwargs, requirement kwargs.
    """
    return s.replace("-", "_")


def convert(name):
    """Converts camel-case strings to snake-case. Used when a user does not define
    a specific name for a registry item.

    Returns
    -------
    val : str
        Name converted to snake-case.
    """
    s1 = re.sub("(.)([A-Z][a-z]+)", r"\1_\2", name)

    return re.sub("([a-z0-9])([A-Z])", r"\1_\2", s1).lower()


class UniqueDictRegistry(DictRegistry):
    """Base registry class that handles hashmap-like associations between a string
    representation of a plugin and the class to be instantiated.
    """
    def add(self, name, cls, overwrite=False):
        """Add an item to the registry.

        Parameters
        ----------
        name : str
            The name referencing the associated class in the registry.
        cls : type
            The class definition (not instance) associated with the name given
            in the first parameter.
        overwrite : bool, optional
            Whether to overwrite an existing entry with the same ``label``.
        """
        if name in self.members and not overwrite:
            raise ValueError(f"Viewer with the name {name} already exists, "
                             f"please choose a different name or pass overwrite=True.")
        else:
            self.members[name] = cls


class ViewerRegistry(UniqueDictRegistry):
    """Registry containing references to custom viewers."""
    def __call__(self, name=None, label=None, overwrite=False):
        def decorator(cls):
            self.add(name, cls, label, overwrite=overwrite)
            return cls
        return decorator

    def add(self, name, cls, label=None, overwrite=False):
        """Add an item to the registry.

        Parameters
        ----------
        name : str
            The key referencing the associated class in the registry
            dictionary.
        cls : type
            The class definition (not instance) associated with the name given
            in the first parameter.
        label : str, optional
            The label displayed in the tooltip when hovering over the tray tab.
        overwrite : bool, optional
            Whether to overwrite an existing entry with the same ``label``.
        """
        if name in self.members and not overwrite:
            raise ValueError(f"Viewer with the name {name} already exists, "
                             f"please choose a different name or pass overwrite=True.")
        else:
            self.members[name] = {'label': label, 'cls': cls}


class TrayRegistry(UniqueDictRegistry):
    """Registry containing references to plugins that will be added to the sidebar
    tray tabs.
    """

    default_viewer_category = [
        "spectrum", "table", "image", "spectrum-2d", "flux", "uncert", "profile"
    ]
    default_viewer_reqs = {
        category: {
            "cls_attr": f"_default_{_to_snake(category)}_viewer_reference_name",
            "init_kwarg": f"{_to_snake(category)}_viewer_reference_name",
            "require_kwargs": [f"require_{_to_snake(category)}_viewer"]
        } for category in default_viewer_category
    }

    def __call__(self, name=None, label=None, icon=None,
                 viewer_requirements=[], overwrite=False):
        def decorator(cls):
            # The class must inherit from `VuetifyTemplate` in order to be
            # ingestible by the component initialization.
            if not issubclass(cls, VuetifyTemplate):
                raise ValueError(
                    f"Unrecognized superclass for {cls.__name__}. All "
                    f"registered components must inherit from "
                    f"`ipyvuetify.VuetifyTemplate`.")

            self.add(name, cls, label, icon, viewer_requirements, overwrite)
            return cls
        return decorator

    def add(self, name, cls, label=None, icon=None,
            viewer_requirements=[], overwrite=False):
        """Add an item to the registry.

        Parameters
        ----------
        name : str
            The key referencing the associated class in the registry
            dictionary.
        cls : type
            The class definition (not instance) associated with the name given
            in the first parameter.
        label : str, optional
            The label displayed in the tooltip when hovering over the tray tab.
        icon : str, optional
            The name of the icon to render in the tray tab.
        viewer_requirements : str, list of str
            Required viewers for this plugin.
        overwrite : bool, optional
            Whether to overwrite an existing entry with the same ``label``.
        """
        if name in self.members and not overwrite:
            raise ValueError(f"Viewer with the name {name} already exists, "
                             f"please choose a different name or pass overwrite=True.")
        else:
            # store the registry name/label so we can access them from the instantiated
            # objects (when determining if a specific plugin is open, for example)
            viewer_reference_name_kwargs = {}

            if not isinstance(viewer_requirements, list):
                viewer_requirements = [viewer_requirements]

            for category in viewer_requirements:
                if category not in self.default_viewer_reqs:
                    raise ValueError(f'Viewer requirements not defined '
                                     f'for viewer category: "{category}" '
                                     f'in plugin "{cls.__class__.__name__}".')

                req = self.default_viewer_reqs[category]
                viewer_cls_attr = req['cls_attr']
                viewer_specific_kwarg = req['init_kwarg']
                requirements = req['require_kwargs']
                viewer_reference_name_kwargs[viewer_cls_attr] = [
                    viewer_specific_kwarg, {
                        k: True for k in requirements
                    }
                ]

            cls._registry_name = name
            cls._registry_label = label
            self.members[name] = {'label': label, 'icon': icon, 'cls': cls,
                                  'viewer_reference_name_kwargs': viewer_reference_name_kwargs}


class ToolRegistry(UniqueDictRegistry):
    """Registry containing references to plugins which will populate the
    application-level toolbar.
    """
    def __call__(self, name=None, overwrite=False):
        def decorator(cls):
            # The class must inherit from `Widget` in order to be
            # ingestible by the component initialization.
            if not issubclass(cls, Widget):
                raise ValueError(
                    f"Unrecognized superclass for `{cls.__name__}`. All "
                    f"registered tools must inherit from "
                    f"`ipywidgets.Widget`.")

            self.add(name, cls, overwrite)
            return cls
        return decorator


class MenuRegistry(UniqueDictRegistry):
    """Registry containing references to plugins that will populate the
    application-level menu bar.
    """
    def __call__(self, name=None, overwrite=False):
        def decorator(cls):
            # The class must inherit from `VuetifyTemplate` in order to be
            # ingestible by the component initialization.
            if not issubclass(cls, Widget):
                raise ValueError(
                    f"Unrecognized superclass for {cls.__name__}. All "
                    f"registered tools must inherit from "
                    f"`ipywidgets.Widget`.")

            self.add(name, cls, overwrite)
            return cls
        return decorator


class DataParserRegistry(UniqueDictRegistry):
    """Registry containing parsing functions for attempting to auto-populate the
    application-defined initial viewers.
    """
    def __call__(self, name=None):
        def decorator(func):
            self.add(name, func)
            return func
        return decorator


class LoaderStepRegistry(UniqueDictRegistry):
    """Registry containing data parsing classes.
    """
    def __init__(self, *args, **kwargs):
        self._step = kwargs.pop('step')
        super().__init__(*args, **kwargs)

    def __call__(self, name=None):
        def decorator(cls):
            cls._registry_label = name
            self.add(name, cls)
            return cls
        return decorator


viewer_registry = ViewerRegistry()
tray_registry = TrayRegistry()
tool_registry = ToolRegistry()
menu_registry = MenuRegistry()
data_parser_registry = DataParserRegistry()  # remove once deconfigging complete
loader_resolver_registry = LoaderStepRegistry(step='resolver')
loader_parser_registry = LoaderStepRegistry(step='parser')
loader_importer_registry = LoaderStepRegistry(step='importer')



import bqplot
import numpy as np
from astropy import units as u
from astropy.coordinates import Angle, SkyCoord
from regions import PolygonSkyRegion
from traitlets import Bool

from jdaviz.core.events import SnackbarMessage
from jdaviz.core.registries import tray_registry
from jdaviz.core.template_mixin import PluginTemplateMixin

__all__ = ['SlitOverlay', 'jwst_header_to_skyregion']


def jwst_header_to_skyregion(header):
    """Convert S_REGION in given FITS header for JWST data into sky region."""
    s_region = header['S_REGION']
    footprint = s_region.split("POLYGON ICRS")[1].split()
    ra = np.array(footprint[::2], dtype=float)
    dec = np.array(footprint[1::2], dtype=float)
    corners = SkyCoord(ra, dec, unit="deg")
    skyregion = PolygonSkyRegion(corners)

    # Need these for zooming
    length = corners[0].separation(corners[1])
    width = corners[1].separation(corners[2])
    skyregion.height = Angle(max(length, width), u.deg)
    skyregion.center = SkyCoord(ra.mean(), dec.mean(), unit="deg")

    return skyregion


@tray_registry('g-slit-overlay', label="Slit Overlay",
               viewer_requirements=['table', 'image', 'spectrum-2d', 'spectrum'])
class SlitOverlay(PluginTemplateMixin):
    template_file = __file__, "slit_overlay.vue"
    visible = Bool(True).tag(sync=True)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # description displayed under plugin title in tray
        self._plugin_description = 'Add a slit to the image viewer.'

        table = self.app.get_viewer(self._default_table_viewer_reference_name)
        table.figure_widget.observe(self.place_slit_overlay, names=['highlighted'])

        self._slit_overlay_mark = None

    @property
    def _default_table_viewer_reference_name(self):
        return getattr(
            self.app._jdaviz_helper, '_default_table_viewer_reference_name', 'table-viewer'
        )

    @property
    def _default_image_viewer_reference_name(self):
        return getattr(
            self.app._jdaviz_helper, '_default_image_viewer_reference_name', 'image-viewer'
        )

    @property
    def _default_spectrum_viewer_reference_name(self):
        return getattr(
            self.app._jdaviz_helper, '_default_spectrum_viewer_reference_name', 'spectrum-viewer'
        )

    @property
    def _default_spectrum_2d_viewer_reference_name(self):
        return getattr(
            self.app._jdaviz_helper,
            '_default_spectrum_2d_viewer_reference_name',
            'spectrum-2d-viewer'
        )

    def vue_change_visible(self, *args, **kwargs):
        if self.visible:
            self.place_slit_overlay()
        else:
            self.remove_slit_overlay()

    def place_slit_overlay(self, *args, **kwargs):
        """
        Find slit information in 2D Spectrum metadata, find the correct
        wcs information from the image metadata, then plot the slit over the
        image viewer using both.
        """
        if not self.visible:
            return

        snackbar_message = None

        # Clear existing slits on the image viewer
        self.remove_slit_overlay()

        # Get data from relevant viewers
        image_data = self.app.get_viewer(
            self._default_image_viewer_reference_name
        ).state.reference_data
        spec2d_data = self.app.get_viewer(
            self._default_spectrum_2d_viewer_reference_name
        ).data()

        # 'S_REGION' contains slit information. Bypass in case no images exist.
        if image_data is not None:
            # Only use S_REGION for Nirspec data, turn the plugin off
            # if other data is loaded
            if (len(spec2d_data) > 0 and 'S_REGION' in spec2d_data[0].meta
                    and spec2d_data[0].meta.get('INSTRUME', '').lower() == "nirspec"):
                header = spec2d_data[0].meta
                sky_region = jwst_header_to_skyregion(header)

                # Use wcs of image viewer to scale slit dimensions correctly
                pix_rec = sky_region.to_pixel(image_data.coords)

                x_coords = pix_rec.vertices.x
                y_coords = pix_rec.vertices.y

                image_viewer = self.app.get_viewer(self._default_image_viewer_reference_name)
                fig_image = image_viewer.figure

                if image_viewer.toolbar.active_tool is not None:
                    image_viewer.toolbar.active_tool = None

                # Create LinearScale that is the same size as the image viewer
                scales = {'x': fig_image.interaction.x_scale, 'y': fig_image.interaction.y_scale}

                # Create slit
                patch2 = bqplot.Lines(x=x_coords, y=y_coords, scales=scales,
                                      fill='none', colors=["red"], stroke_width=2,
                                      close_path=True)

                # Visualize slit on the figure
                fig_image.marks = fig_image.marks + [patch2]

                self._slit_overlay_mark = patch2

            else:
                self.visible = False
                snackbar_message = SnackbarMessage(
                    "\'S_REGION\' not found in Spectrum 2D meta attribute, "
                    "turning slit overlay off",
                    color="warning",
                    sender=self)

        if snackbar_message:
            self.hub.broadcast(snackbar_message)

    def remove_slit_overlay(self):
        if self._slit_overlay_mark is not None:
            image_figure = self.app.get_viewer(
                self._default_image_viewer_reference_name
            ).figure
            # We need to do the following instead of just removing directly on
            # the marks otherwise traitlets doesn't register a change in the
            # marks.
            image_figure.marks.remove(self._slit_overlay_mark)
            image_figure.send_state('marks')
            self._slit_overlay_mark = None
